# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import tkinter as tk

# Third Party Imports
from PIL import Image, ImageTk

# Local Imports
from navigate.tools.image import create_arrow_image


class ArrowLabel(tk.Label):
    """A label that displays an arrow image."""

    def __init__(
        self,
        master,
        *args,
        xys=[],
        direction="right",
        image_width=300,
        image_height=200,
        **kwargs
    ):
        """Initialize the ArrowLabel.

        Parameters
        ----------
        master : tk.Widget
            The parent widget.
        *args : list
            Additional positional arguments to pass to the tk.Label constructor.
        xys : list of tuples
            The coordinates of the arrow.
        direction : str
            The direction of the arrow. One of "right", "left", "up", or "down".
        image_width : int
            The width of the image.
        image_height : int
            The height of the image.
        **kwargs : dict
            Additional keyword arguments to pass to the tk.Label constructor.
        """
        super().__init__(master, *args, **kwargs)
        img = create_arrow_image(xys, image_width, image_height, direction)
        image_gif = img.convert("P", palette=Image.ADAPTIVE)
        #: ImageTk.PhotoImage: The image to display.
        self.image = ImageTk.PhotoImage(image_gif)
        self["image"] = self.image



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import tkinter as tk
from tkinter import messagebox
import inspect
import json
import os

# Third party imports
from PIL import Image, ImageTk

# Local application imports
from navigate.view.popups.feature_list_popup import FeatureIcon, FeatureConfigPopup
from navigate.view.custom_widgets.ArrowLabel import ArrowLabel
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.tools.image import create_arrow_image
from navigate.tools.file_functions import load_yaml_file, save_yaml_file
from navigate.model.features.feature_related_functions import (
    convert_str_to_feature_list,
    convert_feature_list_to_str,
)
from navigate.model.features import feature_related_functions
from navigate.config.config import get_navigate_path


class FeaturePopupController(GUIController):
    """Controller for feature list popup"""

    def __init__(self, view, parent_controller, feature_list_id=0):
        """Initialize the controller

        Parameters
        ----------
        view : navigate.view.popups.feature_list_popup.FeatureListPopup
            The view of the controller
        parent_controller : navigate.controller.main_controller.MainController
            The parent controller
        feature_list_id : int, optional
            The id of the feature list, by default 0
        """
        super().__init__(view, parent_controller)
        #: int: The id of the feature in the feature list.
        self.feature_list_id = feature_list_id
        #: list: The list of feature names.
        self.features = []
        #: list: The list of feature structure.
        self.feature_structure = []

        self.feature_list_graph_controller = FeatureListGraphController(
            self.view.feature_view_frame,
            self.view.inputs["content"],
            self.view.buttons["preview"],
        )

        if "add" in self.view.buttons:
            self.view.buttons["add"].configure(command=self.add_feature_list)
            self.view.popup.protocol("WM_DELETE_WINDOW", self.exit_func)
            self.view.buttons["cancel"].configure(command=self.exit_func)
        elif "confirm" in self.view.buttons:
            self.view.buttons["confirm"].configure(command=self.update_feature_list)
            self.view.popup.protocol("WM_DELETE_WINDOW", self.cancel_acquisition)
            self.view.buttons["cancel"].configure(command=self.cancel_acquisition)

    def populate_feature_list(self, feature_list_id):
        """Populate the feature list

        Parameters
        ----------
        feature_list_id : int
            The id of the feature list
        """
        self.feature_list_id = feature_list_id
        feature_list_content = self.parent_controller.model.get_feature_list(
            feature_list_id
        )
        self.view.inputs["feature_list_name"].set(
            self.parent_controller.menu_controller.feature_list_names[feature_list_id]
        )
        self.view.inputs["feature_list_name"].widget["state"] = "disabled"
        self.feature_list_graph_controller.update(feature_list_content)

    def add_feature_list(self):
        """Add the feature list"""
        if not self.verify_feature_list():
            return
        content = self.view.inputs["content"].get("1.0", "end-1c")
        feature_list_content = "".join(content.split("\n"))
        feature_list_name = self.view.inputs["feature_list_name"].get()
        if not feature_list_name:
            messagebox.showerror(
                title="Feature List Error",
                message="Please enter a name for this feature list!",
            )
            return
        if not self.parent_controller.menu_controller.add_feature_list(
            feature_list_name, feature_list_content
        ):
            messagebox.showerror(
                title="Feature List Error",
                message="Please enter a new list name! "
                "The one you entered has been exist!",
            )
        else:
            self.exit_func()

    def update_feature_list(self):
        """Update the feature list"""
        if not self.verify_feature_list():
            return
        content = self.view.inputs["content"].get("1.0", "end-1c")
        feature_list_content = "".join(content.split("\n"))
        self.parent_controller.execute(
            "load_feature", self.feature_list_id, feature_list_content
        )
        # save to yaml file
        feature_lists_path = get_navigate_path() + "/feature_lists"
        feature_list_name = self.view.inputs["feature_list_name"].get()
        feature_list_config = load_yaml_file(
            os.path.join(
                feature_lists_path, f"{'_'.join(feature_list_name.split(' '))}.yml"
            )
        )
        if feature_list_config and feature_list_config["module_name"] is None:
            save_yaml_file(
                feature_lists_path,
                {
                    "module_name": None,
                    "feature_list_name": feature_list_name,
                    "feature_list": feature_list_content,
                },
                f"{'_'.join(feature_list_name.split(' '))}.yml",
            )
        #: bool: Whether the acquisition should start.
        self.start_acquisiton_flag = True
        self.view.popup.dismiss()

    def verify_feature_list(self):
        """Verify the feature list

        Returns
        -------
        feature_list : list
            The feature list
        """
        content = self.view.inputs["content"].get("1.0", "end-1c")
        return verify_feature_list(content)

    def exit_func(self):
        """Exit the popup"""
        self.view.popup.dismiss()
        delattr(self.parent_controller, "features_popup_controller")

    def cancel_acquisition(self):
        """Cancel the acquisition"""
        self.start_acquisiton_flag = False
        self.view.popup.dismiss()


class FeatureListGraphController:
    def __init__(self, feature_list_view, feature_content_view, preview_btn):
        self.feature_list_view = feature_list_view
        self.feature_content_view = feature_content_view
        self.preview_btn = preview_btn

        self.feature_list = None
        self.features = []
        self.feature_structure = []
        self.feature_list_graph_controllers_true = {}
        self.feature_list_graph_controllers_false = {}

        # get all feature names
        #: list: The list of feature names.
        self.feature_names = []
        temp = dir(feature_related_functions)
        for t in temp:
            if inspect.isclass(getattr(feature_related_functions, t)):
                self.feature_names.append(t)

        # event
        self.preview_btn.configure(command=self.draw_feature_list_graph)

    def update(self, feature_list_content):
        self.feature_list = None
        self.features = []
        self.feature_structure = []
        self.feature_list_graph_controllers_true = {}
        self.feature_list_graph_controllers_false = {}

        if type(feature_list_content) == list:
            feature_list_content = convert_feature_list_to_str(feature_list_content)
        self.feature_content_view.delete("1.0", tk.END)
        self.feature_content_view.insert("1.0", feature_list_content)
        self.draw_feature_list_graph(new_list_flag=True)

    def flatten_feature_list(self, feature_list):
        """Flatten the feature list

        Parameters
        ----------
        feature_list : list
            The feature list

        """
        for temp in feature_list:
            if type(temp) is dict:
                self.features.append(temp)
                self.feature_structure.append(len(self.features) - 1)
            elif temp == "break" or temp == "continue":
                self.features.append(temp)
                self.feature_structure.append(len(self.features) - 1)
            else:
                self.feature_structure.append("(")
                self.flatten_feature_list(temp)
        self.feature_structure.append(")")

    def get_feature_content(self):
        content = self.feature_content_view.get("1.0", "end-1c")
        return content

    def update_feature_content(self):
        self.feature_content_view.delete("1.0", tk.END)
        self.feature_content_view.insert("1.0", self.build_feature_list_text())

    def draw_feature_list_graph(self, new_list_flag=True):
        """Draw the feature list graph

        Parameters
        ----------
        new_list_flag : bool, optional
            Whether the feature list is new, by default True
        """
        if new_list_flag:
            feature_list = verify_feature_list(self.get_feature_content())
            if not feature_list:
                return
            # flatten feature list
            self.features = []
            self.feature_structure = []
            self.flatten_feature_list(feature_list)
            self.feature_structure.pop()

        for child in self.feature_list_view.winfo_children():
            child.destroy()

        l = len(self.features) - 1  # noqa
        feature_icon_width = 228
        al_width = 104
        for i, feature in enumerate(self.features):
            if type(feature) == str:
                btn = FeatureIcon(self.feature_list_view, feature)
                flag = False
            else:
                flag = True
                is_decistion_btn = "true" in feature or "false" in feature
                btn = FeatureIcon(
                    self.feature_list_view,
                    feature["name"].__name__,
                    set_bg=is_decistion_btn,
                )
                # left click
                btn.bind("<Button-1>", self.show_config_popup(i))

            btn.grid(row=0, column=i * 2, sticky="", pady=(30, 0))
            btn["width"] = 20
            # right click
            btn.bind("<Button-3>", self.show_menu(i, flag))
            if i == 0:
                self.feature_list_view.update()
                feature_icon_width = btn.winfo_width()
            if i < l:
                al = ArrowLabel(
                    self.feature_list_view,
                    xys=[(0, 20), (40, 20)],
                    direction="right",
                    image_width=40,
                    image_height=40,
                )
                al.grid(row=0, column=i * 2 + 1, sticky="", pady=(30, 0))
                if i == 0:
                    self.feature_list_view.update()
                    al_width = al.winfo_width()
        # draw loop arrows
        image_width = feature_icon_width * (l + 1) + al_width * l
        image_height = self.calculate_arrow_image_height()
        arrow_height = 50
        stack = []
        arrow_image = None
        space = 30
        end_pos = 0

        for c in self.feature_structure:
            if c == "(":
                if len(stack) > 0 and type(stack[-1]) is int:
                    stack.pop()
                stack.append("(")
            elif c == ")":
                # assert there is already at least one '('
                p, loops, start_pos, arrow_height = stack[-2]
                pre = stack[-1]
                if loops - 1 == 0:
                    stack.pop()
                    stack.pop()
                    stack.append(pre)
                else:
                    stack[-2] = (p, loops - 1, start_pos - space, arrow_height)
                # draw arrow
                arrow_image = create_arrow_image(
                    xys=[
                        (end_pos, 0),
                        (end_pos, arrow_height),
                        (start_pos, arrow_height),
                        (start_pos, 0),
                    ],
                    image_width=image_width,
                    image_height=image_height,
                    direction="up",
                    image=arrow_image,
                )
                # update arrow height
                for i in range(len(stack)):
                    if type(stack[i]) == tuple:
                        p, loops, start_pos, arrow_height = stack[i]
                        stack[i] = (p, loops, start_pos, arrow_height + 20)
            elif type(c) is int:
                if len(stack) > 0 and type(stack[-1]) is int:
                    stack[-1] = c
                else:
                    loops = 0
                    while len(stack) > 0 and stack[-1] == "(":
                        loops += 1
                        stack.pop()
                    if loops > 0:
                        stack.append(
                            (
                                c,
                                loops,
                                c * (feature_icon_width + al_width)
                                + feature_icon_width // 2
                                + (loops // 2) * space,
                                50,
                            )
                        )
                    else:
                        stack.append(c)
                end_pos = c * (feature_icon_width + al_width) + feature_icon_width // 2
        if arrow_image:
            image_gif = arrow_image.convert("P", palette=Image.ADAPTIVE)
            #: ImageTk.PhotoImage: The image of the feature list graph.
            self.image = ImageTk.PhotoImage(image_gif)
            al = tk.Label(self.feature_list_view, image=self.image)
            al.grid(row=1, column=0, columnspan=2 * l + 1, sticky="ew")

    def show_config_popup(self, idx):
        """Show the feature configuration popup

        Parameters
        ----------
        idx : int
            The index of the feature

        Returns
        -------
        func
            The function to show the feature configuration popup
        """
        feature = self.features[idx]

        def func(event):
            """The function to show the feature configuration popup"""
            # load feature parameter setting
            feature_config_path = (
                f"{get_navigate_path()}/"
                f"feature_lists/feature_parameter_setting"
                f"/{feature['name'].__name__}.yml"
            )
            feature_parameter_config = None
            if os.path.exists(feature_config_path):
                feature_parameter_config = load_yaml_file(feature_config_path)
            spec = inspect.getfullargspec(feature["name"])
            if spec.defaults:
                args_value = list(spec.defaults)
            else:
                args_value = spec.defaults
            if "args" in feature:
                for i, a in enumerate(feature["args"]):
                    args_value[i] = a
            kwargs = {}
            if "true" in feature:
                kwargs["true"] = True
            if "false" in feature:
                kwargs["false"] = True
            popup = FeatureConfigPopup(
                self.feature_list_view,
                features=self.feature_names,
                feature_name=feature["name"].__name__,
                args_name=spec.args[2:],
                args_value=args_value,
                title="Feature Parameters",
                parameter_config=feature_parameter_config,
                **kwargs,
            )
            popup.feature_name_widget.widget.bind(
                "<<ComboboxSelected>>", lambda event: refresh_parameters(popup)
            )

            popup.popup.protocol(
                "WM_DELETE_WINDOW", lambda: update_feature_parameters(popup)
            )

            if "true" in feature:
                self.feature_list_graph_controllers_true[
                    idx
                ] = FeatureListGraphController(
                    popup.feature_list_true_frame.feature_view_frame,
                    popup.feature_list_true_frame.content,
                    popup.preview_btn_true,
                )
                self.feature_list_graph_controllers_true[idx].update(feature["true"])
            if "false" in feature:
                self.feature_list_graph_controllers_false[
                    idx
                ] = FeatureListGraphController(
                    popup.feature_list_false_frame.feature_view_frame,
                    popup.feature_list_false_frame.content,
                    popup.preview_btn_false,
                )
                self.feature_list_graph_controllers_false[idx].update(feature["false"])

        def refresh_parameters(popup):
            """Refresh the feature parameters

            Parameters
            ----------
            popup : navigate.view.popups.feature_list_popup.FeatureConfigPopup
                The feature configuration popup
            """
            feature_name = popup.feature_name_widget.get()
            new_feature = getattr(feature_related_functions, feature_name)
            # load feature parameter setting
            feature_config_path = (
                f"{get_navigate_path()}/"
                f"feature_lists/feature_parameter_setting"
                f"/{new_feature.__name__}.yml"
            )
            feature_parameter_config = None
            if os.path.exists(feature_config_path):
                feature_parameter_config = load_yaml_file(feature_config_path)
            spec = inspect.getfullargspec(new_feature)
            popup.build_widgets(spec.args[2:], spec.defaults, feature_parameter_config)

        def update_feature_parameters(popup):
            """Update the feature parameters

            Parameters
            ----------
            popup : navigate.view.popups.feature_list_popup.FeatureConfigPopup
                The feature configuration popup

            Returns
            -------
            func
                The function to update the feature parameters
            """
            widgets = popup.get_widgets()
            feature_name = popup.feature_name_widget.get()
            feature["name"] = getattr(feature_related_functions, feature_name)
            if len(widgets) > 0:
                feature["args"] = [w.get() for w in widgets]
                for i, a in enumerate(feature["args"]):
                    if a == "True":
                        feature["args"][i] = True
                    elif a == "False":
                        feature["args"][i] = False
                    elif popup.inputs_type[i] is float:
                        feature["args"][i] = float(a)
                    elif popup.inputs_type[i] is dict:
                        feature["args"][i] = json.loads(a.replace("'", '"'))
                    elif a == "None":
                        feature["args"][i] = None
            if "true" in feature:
                feature["true"] = convert_str_to_feature_list(
                    self.feature_list_graph_controllers_true[idx].get_feature_content()
                )
            if "false" in feature:
                feature["false"] = convert_str_to_feature_list(
                    self.feature_list_graph_controllers_false[idx].get_feature_content()
                )
            # update text
            self.update_feature_content()
            popup.popup.dismiss()
            self.draw_feature_list_graph(False)

        return func

    def show_menu(self, idx, flag=True):
        """Show the popup menu

        Parameters
        ----------
        idx : int
            The index of the feature

        Returns
        -------
        func
            The function to show the popup menu
        """

        def func(event):
            """The function to show the popup menu"""
            popup_menu = tk.Menu(self.feature_list_view, tearoff=0)
            popup_menu.add_command(label="Delete", command=lambda: delete_feature(idx))
            if flag:
                popup_menu.add_command(
                    label="Insert Before", command=lambda: insert_before(idx)
                )
                popup_menu.add_command(
                    label="Insert After", command=lambda: insert_after(idx)
                )
            popup_menu.post(event.x_root, event.y_root)

        def delete_feature(idx):
            """Delete the feature

            Parameters
            ----------
            idx : int
                The index of the feature
            """
            del self.features[idx]
            i = self.feature_structure.index(idx)
            del self.feature_structure[i]
            for _, c in enumerate(self.feature_structure):
                if type(c) == int and c > idx:
                    self.feature_structure[_] -= 1

            # delete ()
            pre_count = 0
            for _ in range(i - 1, -1, -1):
                if self.feature_structure[_] == "(":
                    del self.feature_structure[_]
                    pre_count += 1
                else:
                    break
            stack = []
            if (
                pre_count == 0
                and i < len(self.feature_structure)
                and self.feature_structure[i] == ")"
            ):
                del self.feature_structure[i]
                for _ in range(i - 1, -1, -1):
                    if self.feature_structure[_] == ")":
                        stack.append(_)
                    elif self.feature_structure[_] == "(":
                        if len(stack) > 0:
                            stack.pop()
                        else:
                            del self.feature_structure[_]
                            break
            else:
                i = i - pre_count
                while pre_count > 0:
                    if self.feature_structure[i] == "(":
                        stack.append(i)
                        i += 1
                    elif self.feature_structure[i] == ")":
                        if len(stack) > 0:
                            stack.pop()
                            i += 1
                        else:
                            del self.feature_structure[i]
                            pre_count -= 1
                    else:
                        i += 1
                    if i >= len(self.feature_structure):
                        break

            # update text
            self.update_feature_content()
            self.draw_feature_list_graph(False)

        def insert_before(idx):
            """Insert the feature before the current feature

            Parameters
            ----------
            idx : int
                The index of the feature
            """
            self.features.insert(idx, dict(self.features[idx]))
            i = self.feature_structure.index(idx)
            for _, c in enumerate(self.feature_structure):
                if type(c) == int and c >= idx:
                    self.feature_structure[_] = c + 1
            self.feature_structure.insert(i, idx)

            # update text
            self.update_feature_content()
            self.draw_feature_list_graph(False)

        def insert_after(idx):
            """Insert the feature after the current feature

            Parameters
            ----------
            idx : int
                The index of the feature

            Returns
            -------
            func
                The function to insert the feature after the current feature
            """
            self.features.insert(
                idx,
                dict(self.features[idx])
                if type(self.features[idx]) == dict
                else self.features[idx],
            )
            i = self.feature_structure.index(idx)
            for _, c in enumerate(self.feature_structure):
                if type(c) == int and c > idx:
                    self.feature_structure[_] = c + 1
            self.feature_structure.insert(i + 1, idx + 1)

            # update text
            self.update_feature_content()
            self.draw_feature_list_graph(False)

        return func

    def calculate_arrow_image_height(self):
        """Calculate the height of the arrow image

        Parameters
        ----------
        feature_structure : list
            The feature structure

        Returns
        -------
        int
            The height of the arrow image
        """
        image_height = 0
        h = 0
        for c in self.feature_structure:
            if c == "(":
                h += 20
                image_height = max(h, image_height)
            elif c == ")":
                h -= 20
        return image_height + 100

    def build_feature_list_text(self):
        """Build the feature list text

        Returns
        -------
        str
            The feature list text
        """
        content = "["
        for c in self.feature_structure:
            if c == "(":
                content += "("
                continue
            elif c == ")":
                content += ")"
            else:
                feature = self.features[c]
                content += "{" + f'"name": {feature["name"].__name__}, '
                if "args" in feature:
                    arg_str = ""
                    for a in feature["args"]:
                        if a is None:
                            arg_str += "None"
                        elif type(a) is bool:
                            arg_str += str(a)
                        elif type(a) is int or type(a) is float:
                            arg_str += str(a)
                        elif type(a) is dict:
                            arg_str += str(a)
                        else:
                            try:
                                float(a)
                                arg_str += a
                            except (ValueError, TypeError):
                                arg_str += f'"{a}"'
                        arg_str += ","
                    content += f'"args": ({arg_str}),'
                # "true"
                if "true" in feature:
                    content += (
                        f'"true": {convert_feature_list_to_str(feature["true"])},'
                    )
                # "false"
                if "false" in feature:
                    content += (
                        f'"false": {convert_feature_list_to_str(feature["false"])},'
                    )
                content += "}"
            content += ","
        content += "]"
        return content


def verify_feature_list(content):
    feature_list_content = "".join(content.split("\n"))
    if feature_list_content in ["break", '"break"', "'break'"]:
        return ["break"]
    if feature_list_content in ["continue", '"continue"', "'continue'"]:
        return ["continue"]
    feature_list = convert_str_to_feature_list(feature_list_content)
    if feature_list is None:
        messagebox.showerror(
            title="Feature List Error",
            message="There is something wrong for this feature "
            "list, please verify there is no "
            "spelling error!",
        )
    return feature_list



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import tkinter as tk
from multiprocessing.managers import ListProxy, DictProxy
import logging

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class StageController(GUIController):
    """StageController

    This class is the controller for the stage GUI. It handles the stage movement
    and the stage limits. It also handles the stage movement buttons and the
    stage movement limits.
    """

    def __init__(self, view, main_view, canvas, parent_controller):
        """Initializes the StageController

        Parameters
         ----------
         view : navigate.view.stage_view.StageView
             The stage view
         main_view : tkinter.Tk
             The main view of the microscope
         canvas : tkinter.Canvas
             The canvas of the microscope
         parent_controller : navigate.controller.Controller
             The parent controller of the stage controller
        """
        super().__init__(view, parent_controller)

        #: str: The default microscope
        self.default_microscope = (
            f"{self.parent_controller.configuration_controller.microscope_name}"
        )

        #: bool: The joystick mode is on.
        self.joystick_is_on = False

        #: list: The joystick axes
        self.joystick_axes = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][self.default_microscope]["stage"].get("joystick_axes", [])

        #: tkinter.Tk: The main view of the microscope
        self.main_view = main_view

        #: tkinter.Canvas: The canvas of the microscope
        self.canvas = canvas

        #: dict: The stage setting dictionary
        self.stage_setting_dict = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]

        #: dict: The event id
        self.event_id = {"x": None, "y": None, "z": None, "theta": None, "f": None}

        # stage movement limits
        #: dict: The minimum stage position.
        self.position_min = {}

        #: dict: The maximum stage position.
        self.position_max = {}

        # variables
        #: dict: The widget values
        self.widget_vals = self.view.get_variables()

        # gui event bind
        buttons = self.view.get_buttons()
        for k in buttons:
            if k[:2] == "up":
                buttons[k].configure(command=self.up_btn_handler(k[3:-4]))
            elif k[:4] == "down":
                buttons[k].configure(command=self.down_btn_handler(k[5:-4]))
            elif k[5:-4] == "xy":
                buttons[k].configure(command=self.xy_zero_btn_handler())
            elif k.startswith("zero"):
                buttons[k].configure(command=self.zero_btn_handler(k[5:-4]))

        for k in ["xy", "z", "f", "theta"]:
            self.widget_vals[k + "_step"].trace_add(
                "write", self.update_step_size_handler(k)
            )

        buttons["stop"].configure(command=self.stop_button_handler)

        buttons["joystick"].configure(command=self.joystick_button_handler)

        #: dict: The position callback traces
        self.position_callback_traces = {}

        #: bool: The position callbacks are bound
        self.position_callbacks_bound = False

        self.bind_position_callbacks()

        #: bool: The stage limits are on
        self.stage_limits = True

        #: list: The stage flip flags
        self.flip_flags = None

        self.initialize()

    def stage_key_press(self, event):
        """The stage key press.

        This function is called when a W, A, S, or D is
        pressed and initializes a stage movement..

        Parameters
        ----------
        event : tkinter.Event
            The tkinter event

        """
        if event.state != 0:
            return
        if not self.joystick_is_on:
            char = event.char.lower()
            if char == "w":
                self.up_btn_handler("y")()
            elif char == "a":
                self.down_btn_handler("x")()
            elif char == "s":
                self.down_btn_handler("y")()
            elif char == "d":
                self.up_btn_handler("x")()

    def initialize(self):
        """Initialize the Stage limits of steps and positions."""
        config = self.parent_controller.configuration_controller
        self.disable_synthetic_stages(config)

        self.position_min = config.get_stage_position_limits("_min")
        self.position_max = config.get_stage_position_limits("_max")

        widgets = self.view.get_widgets()
        step_dict = self.stage_setting_dict[config.microscope_name]
        for axis in ["x", "y", "z", "theta", "f"]:
            widgets[axis].widget.min = self.position_min[axis]
            widgets[axis].widget.max = self.position_max[axis]
            if axis == "x" or axis == "y":
                step_axis = "xy"
            else:
                step_axis = axis
            # the minimum step should be non-zero and non-negative.
            widgets[step_axis + "_step"].widget.configure(from_=1)
            widgets[step_axis + "_step"].widget.configure(to=self.position_max[axis])
            step_increment = step_dict[step_axis + "_step"] // 10
            if step_increment == 0:
                step_increment = 1
            widgets[step_axis + "_step"].widget.configure(increment=step_increment)
            widgets[step_axis + "_step"].set(step_dict[step_axis + "_step"])

        # Joystick
        microscope_name = config.microscope_name

        self.new_joystick_axes = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][f"{microscope_name}"]["stage"].get("joystick_axes", [])

        if self.view.stop_frame.joystick_btn.winfo_ismapped():
            if self.new_joystick_axes is None or list(self.new_joystick_axes) == []:
                self.view.stop_frame.joystick_btn.grid_forget()
            else:
                self.view.stop_frame.joystick_btn.grid()
        else:
            if (
                self.new_joystick_axes is not None
                and list(self.new_joystick_axes) != []
            ):
                self.view.stop_frame.joystick_btn.grid()
            else:
                self.view.stop_frame.joystick_btn.grid_forget()

        if list(self.joystick_axes) != list(self.new_joystick_axes):
            self.force_enable_all_axes()
            self.joystick_is_on = False
            self.new_joystick_axes

        self.joystick_axes = self.new_joystick_axes
        self.flip_flags = config.stage_flip_flags

    def disable_synthetic_stages(self, config):
        """Disable synthetic stages.

        Parameters
        ----------
        config : ConfigurationController
            The configuration controller
        """
        microscope_configuration = config.get_microscope_configuration_dict()
        stages = microscope_configuration["stage"]["hardware"]

        if type(stages) is ListProxy:
            stages = list(stages)
        elif type(stages) is DictProxy:
            stages = [dict(stages)]

        for stage in stages:
            if type(stage) is DictProxy:
                stage_dict = dict(stage)
            elif type(stage) is dict:
                stage_dict = stage

            for axis in stage_dict["axes"]:
                if (stage_dict["type"].lower() == "synthetic") or (
                    stage_dict["type"].lower() == "syntheticstage"
                ):
                    state = "disabled"
                else:
                    state = "normal"

                if axis == "x":
                    self.view.xy_frame.up_x_btn.config(state=state)
                    self.view.xy_frame.down_x_btn.config(state=state)
                    logger.info(f"X axis buttons set to {state}")

                elif axis == "y":
                    self.view.xy_frame.up_y_btn.config(state=state)
                    self.view.xy_frame.down_y_btn.config(state=state)
                    logger.info(f"Y axis buttons set to {state}")

                elif axis == "z":
                    self.view.z_frame.down_btn.config(state=state)
                    self.view.z_frame.up_btn.config(state=state)
                    logger.info(f"Z axis buttons set to {state}")

                elif axis == "theta":
                    self.view.theta_frame.down_btn.config(state=state)
                    self.view.theta_frame.up_btn.config(state=state)
                    logger.info(f"Theta axis buttons set to {state}")

                elif axis == "f":
                    self.view.f_frame.down_btn.config(state=state)
                    self.view.f_frame.up_btn.config(state=state)
                    logger.info(f"Focus axis buttons set to {state}")
            else:
                pass

    def bind_position_callbacks(self):
        """Binds position_callback() to each axis, records the trace name so we can
        unbind later."""
        if not self.position_callbacks_bound:
            for axis in ["x", "y", "z", "theta", "f"]:
                # add event bind to position entry variables
                cbname = self.widget_vals[axis].trace_add(
                    "write", self.position_callback(axis)
                )
                self.position_callback_traces[axis] = cbname
            self.position_callbacks_bound = True

    def unbind_position_callbacks(self):
        """Unbinds position callbacks."""
        if self.position_callbacks_bound:
            for axis, cbname in self.position_callback_traces.items():
                self.widget_vals[axis].trace_remove("write", cbname)
            self.position_callbacks_bound = False

    def populate_experiment_values(self):
        """This function sets all the position and step values"""
        self.stage_setting_dict = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]
        widgets = self.view.get_widgets()
        for k in widgets:
            self.widget_vals[k].set(self.stage_setting_dict.get(k, 0))
            widgets[k].widget.trigger_focusout_validation()

    def set_position(self, position):
        """This function is to populate(set) position in the View

        Parameters
        ----------
        position : dict
            {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
        """
        widgets = self.view.get_widgets()
        for axis in ["x", "y", "z", "theta", "f"]:
            self.widget_vals[axis].set(position.get(axis, 0))
            # validate position value if set through variable
            if self.stage_limits:
                widgets[axis].widget.trigger_focusout_validation()
            self.stage_setting_dict[axis] = position.get(axis, 0)
        self.show_verbose_info("Set stage position")

    def set_position_silent(self, position):
        """This function is to populate(set) position in the View without a trace.

        Parameters
        ----------
        position : dict
            {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
        """
        self.unbind_position_callbacks()

        self.set_position(position)

        self.bind_position_callbacks()

    def get_position(self):
        """This function returns current position from the view.

        Returns
        -------
        position : dict
            Dictionary of x, y, z, theta, and f values.

        """
        position = {}
        try:
            for axis in ["x", "y", "z", "theta", "f"]:
                position[axis] = self.widget_vals[axis].get()
                if self.stage_limits is True:
                    if (
                        position[axis] < self.position_min[axis]
                        or position[axis] > self.position_max[axis]
                    ):
                        return None
        except tk.TclError:
            # Tkinter will raise error when the variable is DoubleVar and the value
            # is empty
            return None
        return position

    def up_btn_handler(self, axis):
        """This function generates command functions according to the desired axis
        to move.

        Parameters
        ----------
        axis : str
            Should be one of 'x', 'y', 'z', 'theta', 'f'
            position_axis += step_axis

        Returns
        -------
        handler : object
            Function for setting desired stage positions in the View.
        """
        position_val = self.widget_vals[axis]
        if axis == "x" or axis == "y":
            step_val = self.widget_vals["xy_step"]
        else:
            step_val = self.widget_vals[axis + "_step"]

        def handler():
            """This function generates command functions according to the desired axis
            to move."""
            stage_direction = -1 if self.flip_flags[axis] else 1
            try:
                temp = position_val.get() + step_val.get() * stage_direction
            except tk._tkinter.TclError:
                return
            if self.stage_limits is True:
                if temp > self.position_max[axis]:
                    temp = self.position_max[axis]
                elif temp < self.position_min[axis]:
                    temp = self.position_min[axis]
            # guarantee stage won't move out of limits
            if position_val.get() != temp:
                position_val.set(temp)

        return handler

    def down_btn_handler(self, axis):
        """This function generates command functions according to the desired axis
        to move.

        Parameters
        ----------
        axis : str
            Should be one of 'x', 'y', 'z', 'theta', 'f'
            position_axis += step_axis

        Returns
        -------
        handler : object
            Function for setting desired stage positions in the View.
        """
        position_val = self.widget_vals[axis]
        if axis == "x" or axis == "y":
            step_val = self.widget_vals["xy_step"]
        else:
            step_val = self.widget_vals[axis + "_step"]

        def handler():
            """This function generates command functions according to the desired axis
            to move."""
            stage_direction = -1 if self.flip_flags[axis] else 1
            try:
                temp = position_val.get() - step_val.get() * stage_direction
            except tk._tkinter.TclError:
                return
            if self.stage_limits is True:
                if temp < self.position_min[axis]:
                    temp = self.position_min[axis]
                elif temp > self.position_max[axis]:
                    temp = self.position_max[axis]
            # guarantee stage won't move out of limits
            if position_val.get() != temp:
                position_val.set(temp)

        return handler

    def zero_btn_handler(self, axis):
        """This function generates command functions according to the desired axis
        to move.

        Parameters
        ----------
        axis : str
            Should be one of 'z', 'theta', 'f'
            position_axis = 0

        Returns
        -------
        handler : object
            Function for setting desired stage positions in the View.
        """
        position_val = self.widget_vals[axis]

        def handler():
            position_val.set(0)

        return handler

    def xy_zero_btn_handler(self):
        """This function generates command functions to set xy position to zero

        Returns
        -------
        handler : object
            Function for setting desired stage positions in the View.
        """
        x_val = self.widget_vals["x"]
        y_val = self.widget_vals["y"]

        def handler():
            x_val.set(0)
            y_val.set(0)

        return handler

    def stop_button_handler(self, *args):
        """This function stops the stage after a 250 ms debouncing period of time."""
        self.view.after(250, lambda *args: self.parent_controller.execute("stop_stage"))

    def joystick_button_handler(self, event=None, *args):
        """Toggle the joystick operation mode.

        Parameters
        ----------
        event : tkinter.Event
            The tkinter event
        """
        if self.joystick_is_on:
            self.joystick_is_on = False
        else:
            self.joystick_is_on = True
        self.view.after(
            250, lambda *args: self.parent_controller.execute("joystick_toggle")
        )
        self.view.toggle_button_states(self.joystick_is_on, self.joystick_axes)

    def force_enable_all_axes(self, event=None, *args):
        """Enables all buttons and entries on the stage tab.

        Parameters
        ----------
        event : tkinter.Event
            The tkinter event (currently unused)
        *args : ...
            ...
        """
        self.view.force_enable_all_axes()

    def position_callback(self, axis, *args, **kwargs):
        """Callback functions bind to position variables.

        Implements debounce functionality for user inputs (or click buttons) to reduce
        time costs of moving stage.

        Parameters
        ----------
        axis : str
            axis can be 'x', 'y', 'z', 'theta', 'f'
        kwargs : ...
            ...

        Returns
        -------
        handler : object
            Function for moving stage to the desired position with debounce
            functionality.
        """
        position_var = self.widget_vals[axis]
        temp = self.view.get_widgets()
        widget = temp[axis].widget

        def handler(*args):
            """Callback functions bind to position variables."""
            if self.event_id[axis]:
                self.view.after_cancel(self.event_id[axis])
            # if position is not a number, then do not move stage
            try:
                position = position_var.get()
                if self.stage_limits:
                    widget.trigger_focusout_validation()
                    # if position is not inside limits do not move stage
                    if (
                        position < float(self.position_min[axis])
                        or position > float(self.position_max[axis])
                    ):
                        return
            except tk._tkinter.TclError:
                if self.event_id[axis]:
                    self.view.after_cancel(self.event_id[axis])
                return
            except AttributeError:
                logger.error(f"Attribute Error Caught: trying to set position {axis}")
                return

            # update stage position
            self.stage_setting_dict[axis] = position
            # Debouncing wait duration - Duration of time to integrate the number of
            # clicks that a user provides. If 1000 ms, if user hits button 10x within
            # 1s, only moves to the final value.
            self.event_id[axis] = self.view.after(
                500,
                lambda *args: self.parent_controller.execute("stage", position, axis),
            )

            self.show_verbose_info("Stage position changed")

        return handler

    def update_step_size_handler(self, axis):
        """Callback functions bind to step size variables

        Parameters
        ----------
        axis : str
            axis can be 'xy', 'z', 'theta', 'f'

        Returns
        -------
        handler : object
            Function to update step size in experiment.yml.
        """

        def func(*args):
            """Callback functions bind to step size variables."""
            microscope_name = self.parent_controller.configuration["experiment"][
                "MicroscopeState"
            ]["microscope_name"]
            try:
                step_size = int(self.widget_vals[axis + "_step"].get())
            except (ValueError, tk._tkinter.TclError):
                return
            self.stage_setting_dict[microscope_name][axis + "_step"] = step_size

        return func



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
import platform
import os
import warnings

# Third Party Imports

# Local Imports
from navigate.controller.controller import Controller
from navigate.log_files.log_functions import log_setup
from navigate.view.splash_screen import SplashScreen
from navigate.tools.main_functions import (
    evaluate_parser_input_arguments,
    create_parser,
)

# Proxy Configuration
os.environ["http_proxy"] = ""
os.environ["https_proxy"] = ""


def main():
    """Light-sheet Microscopy (Navigate).
    Microscope control software built in a Model-View-Controller architecture.
    Provides control of cameras, data acquisition cards, filter wheels, lasers
    stages, voice coils, and zoom servos.

    Parameters
    ----------
    *args : iterable
        --synthetic-hardware
        --sh
        --config-file
        --experiment-file
        --waveform_constants-path
        --rest-api-file
        --waveform-templates-file
        --logging-config

    Returns
    -------
    None

    Examples
    --------
    >>> python main.py --synthetic-hardware
    """
    if platform.system() != 'Windows':
        print("WARNING: navigate was built to operate on a Windows platform. "
              "While much of the software will work for evaluation purposes, some "
              "unanticipated behaviors may occur. For example, it is known that the "
              "Tkinter-based GUI does not grid symmetrically, nor resize properly "
              "on MacOS. Testing on Linux operating systems has not been performed.")

    # Start the GUI, withdraw main screen, and show splash screen.
    root = tk.Tk()
    root.withdraw()

    # Splash Screen
    current_directory = os.path.dirname(os.path.realpath(__file__))
    splash_screen = SplashScreen(
        root, os.path.join(current_directory, "view", "icon", "splash_screen_image.png")
    )

    # Parse command line arguments
    parser = create_parser()
    args = parser.parse_args()

    (
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
        logging_path,
    ) = evaluate_parser_input_arguments(args)

    log_setup("logging.yml", logging_path)

    Controller(
        root,
        splash_screen,
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
        args,
    )
    root.mainloop()


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
from queue import Queue
import threading

# Third Party Imports
import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import linregress

# Local imports
from navigate.model.features.feature_container import load_features
from navigate.model.analysis.image_contrast import fast_normalized_dct_shannon_entropy


def power_tent(x, x_offset, y_offset, amplitude, sigma, alpha):
    """Power tent function.

    Used for fitting the response curve of the autofocus routine.

    Parameters
    ----------
    x : float
        x value
    x_offset : float
        x offset
    y_offset : float
        y offset
    amplitude : float
        amplitude
    sigma : float
        sigma
    alpha : float
        alpha

    Returns
    -------
    function : float
        Power tent function
    """
    function = y_offset + amplitude * (1 - np.abs(sigma * (x - x_offset)) ** alpha)
    return function


class Autofocus:
    """Autofocus Data Process

    This function is called by the data thread. It will get the data from the
    autofocus_frame_queue and calculate the entropy of the image. The entropy
    is then compared to the maximum entropy and the position is saved if it is
    higher. The autofocus_pos_queue is then filled with the next position to
    move to. If the autofocus_pos_queue is empty, the autofocus is finished.
    """

    def __init__(self, model, device="stage", device_ref="f"):
        """Initialize the Autofocus class.

        Parameters
        ----------
        model : Model
            Model object
        device : str
            Device name
        device_ref : str
            Device reference
        """
        #: Model: Model object
        self.model = model
        #: float: Maximum entropy
        self.max_entropy = None
        #: int: Frame id
        self.f_frame_id = None
        #: int: Number of frames
        self.frame_num = None
        #: float: Initial position
        self.init_pos = None
        #: float: Focus position
        self.f_pos = None
        #: float: Focus position
        self.focus_pos = None
        #: int: Target frame id
        self.target_frame_id = None
        #: int: Number of frames
        self.get_frames_num = None
        #: list: Plot data
        self.plot_data = None
        #: int: Total frame number
        self.total_frame_num = None
        #: float: Fine step size
        self.fine_step_size = None
        #: float: Fine position offset
        self.fine_pos_offset = None
        #: float: Coarse step size
        self.coarse_step_size = None
        #: int: Coarse steps
        self.coarse_steps = None
        #: int: Signal id
        self.signal_id = None

        #: Queue: Autofocus frame queue
        self.autofocus_frame_queue = Queue()
        #: Queue: Autofocus position queue
        self.autofocus_pos_queue = Queue()

        #: int: Target channel
        self.target_channel = 1
        #: dict: Configuration table
        self.config_table = {
            "signal": {
                "init": self.pre_func_signal,
                "main": self.in_func_signal,
                "end": self.end_func_signal,
            },
            "data": {
                "init": self.pre_func_data,
                "main": self.in_func_data,
                "end": self.end_func_data,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }
        #: str: Device name
        self.device = device
        #: str: Device reference
        self.device_ref = device_ref

    def run(self):
        """Run the Autofocusing Routine

        Returns
        -------
        dict
            Autofocus parameters.
        """
        frame_num = self.get_autofocus_frame_num()
        if frame_num < 1:
            return

        # Opens correct shutter and puts all signals to false
        self.model.prepare_acquisition()
        self.model.active_microscope.prepare_next_channel()

        # load Autofocus
        self.model.signal_container, self.model.data_container = load_features(
            self.model,
            [
                [
                    {
                        "name": Autofocus,
                        "args": (
                            self.device,
                            self.device_ref,
                        ),
                    }
                ]
            ],
        )

        self.model.signal_thread = threading.Thread(
            target=self.model.run_acquisition,
            name="Autofocus Signal",
        )

        self.model.data_thread = threading.Thread(
            target=self.model.run_data_process,
            args=(frame_num + 1,),
            name="Autofocus Data",
        )

        # Start Threads
        self.model.signal_thread.start()
        self.model.data_thread.start()

    def get_autofocus_frame_num(self):
        """Calculate how many frames are needed to get the best focus position.

        Returns
        -------
        int
            Number of frames to be processed.
        """
        settings = self.model.configuration["experiment"]["AutoFocusParameters"][
            self.model.active_microscope_name
        ][self.device][self.device_ref]
        frames = 0
        if settings["coarse_selected"]:
            coarse_range = float(settings["coarse_range"])
            coarse_step_size = float(settings["coarse_step_size"])
            frames = int(coarse_range // coarse_step_size) + 1
        if settings["fine_selected"]:
            fine_range = float(settings["fine_range"])
            fine_step_size = float(settings["fine_step_size"])
            frames += int(fine_range // fine_step_size) + 1
        return frames

    @staticmethod
    def get_steps(ranges, step_size):
        """Calculate number of steps for autofocusing routine.

        Parameters
        ----------
        ranges : float
            Distance to be traveled during the autofocusing routine
        step_size : float
            Step size for autofocusing routine

        Returns
        -------
        steps : float
            Number of steps for the stack
        pos_offset : float
            Need to figure out.
        """
        steps = ranges // step_size + 1
        pos_offset = (steps // 2) * step_size + step_size
        return steps, pos_offset

    def pre_func_signal(self):
        """Prepare the autofocus routine."""
        settings = self.model.configuration["experiment"]["AutoFocusParameters"][
            self.model.active_microscope_name
        ][self.device][self.device_ref]
        if self.device == "stage":
            self.focus_pos = self.model.configuration["experiment"]["StageParameters"][
                self.device_ref
            ]
        else:
            self.focus_pos = 0
        self.total_frame_num = self.get_autofocus_frame_num()  # Total frame num
        self.coarse_steps, self.init_pos = 0, 0

        if settings["fine_selected"]:
            self.fine_step_size = float(settings["fine_step_size"])
            fine_steps, self.fine_pos_offset = self.get_steps(
                float(settings["fine_range"]), self.fine_step_size
            )
            self.init_pos = self.focus_pos - self.fine_pos_offset

        if settings["coarse_selected"]:
            self.coarse_step_size = float(settings["coarse_step_size"])
            self.coarse_steps, coarse_pos_offset = self.get_steps(
                float(settings["coarse_range"]), self.coarse_step_size
            )
            self.init_pos = self.focus_pos - coarse_pos_offset
        self.signal_id = 0

    def in_func_signal(self):
        """Run the autofocus routine."""

        if self.signal_id < self.coarse_steps:
            self.init_pos += self.coarse_step_size
            if self.device == "stage":
                self.model.move_stage(
                    {f"{self.device_ref}_abs": self.init_pos}, wait_until_done=True
                )
                self.model.logger.debug(
                    f"*** Autofocus move stage: ({self.device_ref}, {self.init_pos})"
                )
            elif self.device == "remote_focus":
                self.model.active_microscope.move_remote_focus(self.init_pos)
                self.model.logger.debug(
                    f"*** Autofocus move remote focus: {self.init_pos}"
                )
            self.autofocus_frame_queue.put(
                (self.model.frame_id, self.coarse_steps - self.signal_id, self.init_pos)
            )

        elif self.signal_id < self.total_frame_num:
            if self.signal_id and self.signal_id == self.coarse_steps:
                self.init_pos = self.autofocus_pos_queue.get(
                    timeout=self.coarse_steps * 10
                )
                self.init_pos -= self.fine_pos_offset
            self.init_pos += self.fine_step_size
            if self.device == "stage":
                self.model.move_stage(
                    {f"{self.device_ref}_abs": self.init_pos}, wait_until_done=True
                )
                self.model.logger.debug(
                    f"*** Autofocus move stage: ({self.device_ref}, {self.init_pos})"
                )
            elif self.device == "remote_focus":
                self.model.active_microscope.move_remote_focus(self.init_pos)
                self.model.logger.debug(
                    f"*** Autofocus move remote focus: {self.init_pos}"
                )
            self.autofocus_frame_queue.put(
                (
                    self.model.frame_id,
                    self.total_frame_num - self.signal_id,
                    self.init_pos,
                )
            )

        else:
            self.init_pos = self.autofocus_pos_queue.get(timeout=self.coarse_steps * 10)
            if self.device == "stage":
                self.model.move_stage(
                    {f"{self.device_ref}_abs": self.init_pos}, wait_until_done=True
                )
                self.model.logger.debug(
                    f"*** Autofocus move stage: ({self.device_ref}, {self.init_pos})"
                )
            elif self.device == "remote_focus":
                self.model.active_microscope.move_remote_focus(self.init_pos)
                self.model.logger.debug(
                    f"*** Autofocus move remote focus: {self.init_pos}"
                )

        self.signal_id += 1
        return self.init_pos if self.signal_id > self.total_frame_num else None

    def end_func_signal(self):
        """End the autofocus routine."""

        return self.signal_id > self.total_frame_num

    def pre_func_data(self):
        """Prepare the autofocus routine."""
        # Initialize the autofocus data
        self.max_entropy = 0
        self.f_frame_id = -1

        # Need to calculate DCTS value, but the image frame isn't ready
        self.frame_num = 10  # any value but not 1
        self.f_pos = 0
        self.target_frame_id = 0  # frame id in the buffer with best focus
        self.get_frames_num = 0
        self.plot_data = []
        self.total_frame_num = self.get_autofocus_frame_num()

    def in_func_data(self, frame_ids=[]):
        """Run the autofocus routine.

        Parameters
        ----------
        frame_ids : list
            List of frame ids to be processed.

        Returns
        -------
        frame_ids : list
            List of frame ids to be processed.
        """

        self.get_frames_num += len(frame_ids)
        while True:
            try:
                if self.f_frame_id < 0:
                    (
                        self.f_frame_id,
                        self.frame_num,
                        self.f_pos,
                    ) = self.autofocus_frame_queue.get_nowait()
                if self.f_frame_id not in frame_ids:
                    break
            except Exception:
                break

            entropy = fast_normalized_dct_shannon_entropy(
                input_array=self.model.data_buffer[self.f_frame_id],
                psf_support_diameter_xy=3,
            )

            self.model.logger.debug(
                f"Appending plot data for frame {self.f_frame_id} focus: {self.f_pos}, "
                f"entropy: {entropy[0]}"
            )
            self.plot_data.append([self.f_pos, entropy[0]])
            # Need to initialize entropy above for the first iteration of the autofocus
            # routine. Need to initialize entropy_vector above for the first iteration
            # of the autofocus routine. Then need to append each measurement to the
            # entropy_vector.  First column will be the focus position, second column
            # would be the DCT entropy value.

            # Find Maximum Focus Position
            if entropy > self.max_entropy:
                self.max_entropy = entropy
                self.focus_pos = self.f_pos
                self.target_frame_id = self.f_frame_id

            self.f_frame_id = -1

            if self.frame_num == 1:
                self.frame_num = 10  # any value but not 1
                self.model.logger.info(
                    f"***********max shannon entropy: {self.max_entropy}, "
                    f"{self.focus_pos}"
                )
                # find out the focus
                self.autofocus_pos_queue.put(self.focus_pos)
                # return [self.target_frame_id]

        if self.get_frames_num > self.total_frame_num:
            return frame_ids

    def end_func_data(self):
        """End the autofocus routine.

        Returns
        -------
        bool
            True if the autofocus routine is finished.
        """
        if self.get_frames_num <= self.total_frame_num:
            return False

        # Send the data for plotting via the event queue
        self.model.event_queue.put(("autofocus", [self.plot_data, False, True]))

        # Evaluate data by fitting it to an inverse power tent.
        if self.model.configuration["experiment"]["AutoFocusParameters"][
            self.model.active_microscope_name
        ][self.device][self.device_ref]["robust_fit"]:
            fit_data, fit_focus_position, r_squared = self.robust_autofocus()

            # If the fit is good, use the fit focus position, else use the max entropy
            if r_squared > 0.9:
                self.focus_pos = fit_focus_position
                self.model.event_queue.put(("autofocus", [fit_data, True, False]))
                self.model.logger.info(
                    f"Robust Focus Estimate: {self.focus_pos}, " f"R^2: {r_squared}"
                )
            else:
                print("Robust Focus Estimate Failed. R^2: %s" % r_squared)
                self.model.logger.info(
                    f"Robust Focus Estimate Failed. R^2: {r_squared}"
                )

        # Update the configuration with the new focus position
        if self.device == "stage":
            self.model.configuration["experiment"]["StageParameters"][
                self.device_ref
            ] = self.focus_pos

            # Tell the controller to update the view
            stage_position = dict(
                map(
                    lambda axis: (
                        f"{axis}_abs",
                        self.model.configuration["experiment"]["StageParameters"][axis],
                    ),
                    ["x", "y", "z", "f", "theta"],
                )
            )
            self.model.event_queue.put(("update_stage", stage_position))
        elif self.device == "remote_focus":
            # update offset of the waveform_constants configuration dict
            zoom = self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
            remote_focus_constants = self.model.configuration["waveform_constants"][
                "remote_focus_constants"
            ][self.model.active_microscope_name][zoom]
            for laser in remote_focus_constants.keys():
                remote_focus_constants[laser]["offset"] = (
                    float(remote_focus_constants[laser]["offset"]) + self.focus_pos
                )

        # Log the new focus position
        # self.model.logger.info("***********final focus: %s" % self.focus_pos)
        # self.model.logger.info(
        #     f"***** final stage position: {self.model.get_stage_position()}"
        # )
        return self.get_frames_num > self.total_frame_num

    def robust_autofocus(self):
        """Robust autofocus routine.

        TODO: Current values for amplitude, sigma, and alpha are hard-coded. Fitting
        is unfortunately unstable.

        Returns
        -------
        fit_data : list
            Autofocus data
        focus_position : float
            Focus position
        r_squared : float
            R-Squared value
        """

        # Convert plot data to numpy array
        x_data = np.asarray(self.plot_data)[:, 0]
        y_data = np.asarray(self.plot_data)[:, 1]

        # Provide starting conditions for the fit.
        x_offset = np.argmax(y_data)
        y_offset = np.min(y_data)
        amplitude = 1
        sigma = 1
        alpha = 0.01

        # Fit the data
        start_vals = [x_offset, y_offset, amplitude, sigma, alpha]
        tent, _ = curve_fit(power_tent, x_data, y_data, p0=start_vals)
        y_fit = power_tent(x_data, *tent)
        focus_value = np.argmax(y_fit)
        focus_position = x_data[focus_value]

        # Calculate the R-Squared value
        _, _, r_value, _, _ = linregress(y_data, y_fit)
        r_squared = r_value**2

        fit_data = []
        for i in range(len(x_data)):
            fit_data.append([x_data[i], y_fit[i]])

        return fit_data, focus_position, r_squared



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os

# Third Party Imports
import h5py
import zarr  # for n5
import numpy as np
import numpy.typing as npt

# Local imports
from .data_source import DataSource
from ..metadata_sources.bdv_metadata import BigDataViewerMetadata
from multiprocessing.managers import DictProxy


class BigDataViewerDataSource(DataSource):
    """BigDataViewer data source.

    This class is used to write data to a BigDataViewer-compatible file. It
    supports both HDF5 and N5 file formats. The file is written in a
    multi-resolution pyramid format, with each resolution level subdivided into
    32x32x1 blocks. The number of blocks in each dimension is determined by the
    shape of the data and the resolution level.
    """

    def __init__(self, file_name: str = None, mode: str = "w") -> None:
        """Initializes the BigDataViewerDataSource.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        mode : str
            The mode to open the file in. Must be "w" for write or "r" for read.
        """
        #: np.array: The resolution of each down-sampled pyramid level.
        self._resolutions = np.array(
            [[1, 1, 1], [2, 2, 1], [4, 4, 1], [8, 8, 1]], dtype=int
        )
        #: np.array: The number of subdivisions in each dimension.
        self._subdivisions = None
        #: np.array: The shape of the image.
        self._shapes = None
        #: np.array: The image.
        self.image = None
        #: list: The views.
        self._views = []
        #: zarr.N5Store: The N5 store.
        self.__store = None
        #: str: The file type.
        self.__file_type = os.path.splitext(os.path.basename(file_name))[-1][1:].lower()
        if self.__file_type not in ["h5", "n5"]:
            raise ValueError(f"Unknown file type {self.__file_type}.")
        if self.__file_type == "h5":
            self.setup = self._setup_h5
            self.ds_name = self._h5_ds_name
        elif self.__file_type == "n5":
            self.setup = self._setup_n5
            self.ds_name = self._n5_ds_name

        # self._current_frame = 0
        #: BigDataViewerMetadata: The metadata.
        self.metadata = BigDataViewerMetadata()

        super().__init__(file_name, mode)

    def __getitem__(self, keys):
        """Magic method to get slice requests passed by, e.g., ds[:,2:3,...].
        Allows arbitrary slicing of dataset via calls to get_slice().

        Order is xycztps where x, y, z are Cartesian indices, c is channel,
        t is timepoints, p is positions and s is subdivisions to index along.

        TODO: Add subdivisions.

        Parameters
        ----------
        keys : tuple
            Tuple of indices.

        Returns
        -------
        npt.ArrayLike
            Array of shape (p, t, z, c, y, x)
        """

        # Check lengths
        if isinstance(keys, slice) or isinstance(keys, int):
            length = 1
        else:
            length = len(keys)
        if length < 1:
            raise IndexError(
                "Too few indices. Indices may be (x, y, c, z, t, p, subdiv)."
            )
        elif length > 7:
            raise IndexError(
                "Too many indices. Indices may be (x, y, c, z, t, p, subdiv)."
            )

        # Handle "slice the rest"
        if length > 1 and keys[-1] == Ellipsis:
            keys = keys[:-2]
            length -= 1

        def ensure_iter(pos):
            """Ensure the input is iterable.

            Parameters
            ----------
            pos : int
                The position.

            Returns
            -------
            range
                The range.
            """
            if length > pos:
                try:
                    val = keys[pos]
                except TypeError:
                    # Only one key
                    val = keys
                if isinstance(val, slice):
                    if val.start is None and val.stop is None and val.step is None:
                        return range(self.shape[pos])
                    return range(10**10)[val]
                elif isinstance(val, int):
                    return range(val, val + 1)
            else:
                return range(self.shape[pos])

        def ensure_slice(pos):
            """Ensure the input is a slice or a single integer.

            Parameters
            ----------
            pos : int
                The position.

            Returns
            -------
            slice
                The slice.
            """
            # TODO: Handle list as input
            if length > pos:
                try:
                    val = keys[pos]
                except TypeError:
                    # Only one key
                    val = keys
                assert isinstance(val, slice) or isinstance(val, int)
                return val
            else:
                # Default to all values
                return slice(None, None, None)

        # Get legal indices
        xs = ensure_slice(0)
        ys = ensure_slice(1)
        cs = ensure_iter(2)
        zs = ensure_slice(3)
        ts = ensure_iter(4)
        if length > 5:
            val = keys[5]
            if isinstance(val, slice):
                if val.start is None and val.stop is None and val.step is None:
                    ps = range(self.positions)
                else:
                    ps = range(10**10)[val]
            elif isinstance(val, int):
                ps = range(val, val + 1)
        else:
            ps = range(self.positions)

        if length > 6 and isinstance(keys[6], int):
            subdiv = keys[6]
        else:
            subdiv = 0

        if len(cs) == 1 and len(ts) == 1 and len(ps) == 1:
            return self.get_slice(xs, ys, cs[0], zs, ts[0], ps[0], subdiv)

        def slice_len(sl, n):
            """Calculate the length of the slice over an array of size n.

            Parameters
            ----------
            sl : slice
                The slice.
            n : int
                The size of the array.

            Returns
            -------
            int
                The length of the slice.
            """
            sx = sl.indices(n)
            return (sx[1] - sx[0]) // sx[2]

        sliced_ds = np.empty(
            (
                len(ps),
                len(ts),
                slice_len(zs, self.shape_z) // self.resolutions[subdiv][2],
                len(cs),
                slice_len(ys, self.shape_y) // self.resolutions[subdiv][1],
                slice_len(xs, self.shape_x) // self.resolutions[subdiv][0],
            ),
            dtype=np.uint16,
        )

        for c in cs:
            for t in ts:
                for p in ps:
                    sliced_ds[p, t, :, c, :, :] = self.get_slice(
                        xs, ys, c, zs, t, p, subdiv
                    )

        return sliced_ds

    def get_slice(self, x, y, c, z=0, t=0, p=0, subdiv=0):
        """Get a single slice of the dataset.

        Parameters
        ----------
        x : int or slice
            x indices to grab
        y : int or slice
            y indices to grab
        c : int
            Single channel
        z : int or slice
            z indices to grab
        t : int
            Single timepoint
        p : int
            Single position
        subdiv : int
            Subdivision of the dataset to index along

        Returns
        -------
        npt.ArrayLike
            3D (z, y, x) slice of data set
        """
        setup = self.ds_name(t, c, p).replace("???", str(subdiv))
        return self.image[setup][z, y, x]

    @property
    def resolutions(self) -> npt.ArrayLike:
        """Getter for resolutions.

        Store as XYZ per BDV spec.

        Returns
        -------
        resolutions : npt.ArrayLike
            The resolutions.
        """
        return self._resolutions

    @property
    def subdivisions(self) -> npt.ArrayLike:
        """Getter for subdivisions.

        Store as XYZ per BDV spec.

        Returns
        -------
        subdivisions : npt.ArrayLike
            The subdivisions.
        """
        if self._subdivisions is None:
            self._subdivisions = np.zeros((4, 3), dtype=int)
            self._subdivisions[:, 0] = np.gcd(32, self.shapes[:, 0])
            self._subdivisions[:, 1] = np.gcd(32, self.shapes[:, 1])
            self._subdivisions[:, 2] = np.gcd(32, self.shapes[:, 2])

            # Safety
            self._subdivisions = np.maximum(self._subdivisions, 1)

            # Reverse to XYZ
            self._subdivisions = self._subdivisions[:, ::-1]
        return self._subdivisions

    @property
    def shapes(self) -> npt.ArrayLike:
        """Getter for image shape.

        Store as ZYX rather than XYZ, per BDV spec.

        Returns
        -------
        shapes : npt.ArrayLike
            The shapes.
        """
        if self._shapes is None:
            self._shapes = np.maximum(
                np.ceil(
                    np.array([self.shape_z, self.shape_y, self.shape_x])[None, :]
                    / self.resolutions[:, ::-1]
                ).astype(int),
                1,
            )
        return self._shapes

    @property
    def nbytes(self) -> int:
        """Getter for image size.

        Size in bytes. Overrides base class. Accounts for subdivisions.

        Returns
        -------
        size : int
            The size of the image in bytes.
        """
        return (
            np.prod(self.shapes, axis=1)
            * self.shape_t
            * self.shape_c
            * self.positions
            * self.bits
            // 8
        ).sum()

    def set_metadata_from_configuration_experiment(
        self, configuration: DictProxy
    ) -> None:
        """Sets the metadata from according to the microscope configuration.

        Parameters
        ----------
        configuration : DictProxy
            The configuration experiment.
        """
        self._subdivisions = None
        self._shapes = None

        # Set rotation and affine transform information in metadata.
        self.metadata.get_affine_parameters(configuration=configuration)
        return super().set_metadata_from_configuration_experiment(configuration)

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Writes data to the image file.

        Parameters
        ----------
        data : npt.ArrayLike
            The data to write.
        kw : dict
            The keyword arguments to write.
        """
        self.mode = "w"

        is_kw = len(kw) > 0

        c, z, t, p = self._cztp_indices(
            self._current_frame, self.metadata.per_stack
        )  # find current channel

        if not (z or c or t or p):
            self.setup()

        ds_name = self.ds_name(t, c, p)
        for i in range(self.subdivisions.shape[0]):
            dx, dy, dz = self.resolutions[i, ...]
            if z % dz == 0:
                dataset_name = ds_name.replace("???", str(i))
                # print(z, dz, dataset_name, self.image[dataset_name].shape,
                #       data[::dx, ::dy].shape)
                zs = min(z // dz, self.shapes[i, 0] - 1)  # TODO: Is this necessary?
                self.image[dataset_name][zs, ...] = data[::dy, ::dx].astype(np.int16)
                if is_kw and (i == 0):
                    self._views.append(kw)
        self._current_frame += 1

        # Check if this was the last frame to write
        c, z, t, p = self._cztp_indices(self._current_frame, self.metadata.per_stack)
        if (z == 0) and (c == 0) and ((t >= self.shape_t) or (p >= self.positions)):
            self.setup(
                self.shape_c * self.positions, self.shape_c * (p + 1), create_flag=False
            )
            self.positions = p + 1

    def _h5_ds_name(self, t, c, p):
        """Get the HDF5 dataset name for the given timepoint, channel, and position.

        Parameters
        ----------
        t : int
            The timepoint.
        c : int
            The channel.
        p : int
            The position.

        Returns
        -------
        ds_name : str
            The dataset name.
        """
        time_group_name = f"t{t:05}"
        setup_group_name = f"s{(c*self.positions+p):02}"
        ds_name = "/".join([time_group_name, setup_group_name, "???", "cells"])
        return ds_name

    def _n5_ds_name(self, t, c, p):
        """Get the N5 dataset name for the given timepoint, channel, and position.

        Parameters
        ----------
        t : int
            The timepoint.
        c : int
            The channel.
        p : int
            The position.

        Returns
        -------
        ds_name : str
            The dataset name.
        """
        time_group_name = f"timepoint{t}"
        setup_group_name = f"setup{(c*self.positions+p)}"
        ds_name = "/".join([setup_group_name, time_group_name, "s???"])
        return ds_name

    def read(self) -> None:
        """Reads data from the image file."""
        self.mode = "r"
        if self.__file_type == "h5":
            self.image = h5py.File(self.file_name, "r")
        elif self.__file_type == "n5":
            self.image = zarr.N5Store(self.file_name)
        xml_fn = os.path.splitext(self.file_name)[0] + ".xml"
        self.metadata.parse_xml(xml_fn)
        self.get_shape_from_metadata()

    def _setup_h5(self, *args, create_flag=True):
        """Set up the HDF5 file.

        This function creates the file and the datasets to populate.
        """
        if create_flag:
            self.image = h5py.File(self.file_name, "a")

        setup_start, setup_end = 0, self.shape_c * self.positions
        if len(args) >= 2:
            setup_start, setup_end = args[0], args[1]

        # Create setups
        for i in range(setup_start, setup_end):
            setup_group_name = f"s{i:02}"
            if setup_group_name in self.image:
                del self.image[setup_group_name]
            self.image.create_dataset(
                f"{setup_group_name}/resolutions",
                data=self.resolutions,
                dtype="float64",
            )
            self.image.create_dataset(
                f"{setup_group_name}/subdivisions",
                data=self.subdivisions,
                dtype="int32",
            )

        # Create the datasets to populate
        for t in range(self.shape_t):
            time_group_name = f"t{t:05}"
            for i in range(setup_start, setup_end):
                setup_group_name = f"s{i:02}"
                for j in range(self.subdivisions.shape[0]):
                    dataset_name = "/".join(
                        [time_group_name, setup_group_name, f"{j}", "cells"]
                    )
                    if dataset_name in self.image:
                        del self.image[dataset_name]
                    # print(f"Creating {dataset_name} with shape {self.shapes[j,...]}")
                    self.image.create_dataset(
                        dataset_name,
                        chunks=tuple(self.subdivisions[j, ...][::-1]),
                        shape=self.shapes[j, ...],
                        dtype="int16",
                    )

    def _setup_n5(self, *args, create_flag=True):
        """Set up the N5 file.

        This function creates the file and the datasets to populate. By default,
        it appears to implement blosc compression. Consequently, the anticipated file
        size, and the actual file size, do not match. This is not the case for HDF5.

        Note
        ----
            Setups and timepoints are flipped in N5 vs. HDF5, see
            https://github.com/bigdataviewer/bigdataviewer-core/blob/master/BDV%20N5%20format.md

        """
        if create_flag:
            self.__store = zarr.N5Store(self.file_name)
            self.image = zarr.group(store=self.__store, overwrite=True)

        setup_start, setup_end = 0, self.shape_c * self.positions
        if len(args) >= 2:
            setup_start, setup_end = args[0], args[1]

        for i in range(setup_start, setup_end):
            setup_group_name = f"setup{i}"
            setup = self.image.create_group(setup_group_name)
            setup.attrs["downsamplingFactors"] = self.resolutions.tolist()
            setup.attrs["dataType"] = "int16"
            for t in range(self.shape_t):
                time_group_name = f"timepoint{t}"
                timepoint = setup.create_group(time_group_name)
                for j in range(self.subdivisions.shape[0]):
                    s_group_name = f"s{j}"
                    shape = [int(x) for x in self.shapes[j, ...][::-1]]
                    # chunks = self.subdivisions[j, ...]
                    sx = timepoint.zeros(
                        s_group_name, shape=tuple(shape), chunks=(shape[0], shape[1], 1)
                    )
                    sx.attrs["dataType"] = "int16"
                    sx.attrs["blockSize"] = [shape[0], shape[1], 1]
                    sx.attrs["dimensions"] = list(shape)
        # print(self.image.tree())

    def _mode_checks(self) -> None:
        """Checks that the mode is valid."""
        self._write_mode = self._mode == "w"
        self.close()  # if anything was already open, close it
        if self._write_mode:
            self._current_frame = 0
            self._views = []
            self.setup()
        else:
            self.read()
        self._closed = False

    def close(self) -> None:
        """Close the image file."""
        if self._closed:
            return
        self._check_shape(self._current_frame - 1, self.metadata.per_stack)
        if self.__file_type == "n5":
            self.__store.close()
        else:
            self.image.close()
        if self.mode != "r":
            self.metadata.write_xml(self.file_name, views=self._views)
        self._closed = True



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import os
from pathlib import Path
import inspect

from navigate.tools.common_functions import load_module_from_file
from navigate.tools.file_functions import load_yaml_file, save_yaml_file
from navigate.tools.decorators import FeatureList, AcquisitionMode
from navigate.model.features import feature_related_functions
from navigate.config.config import get_navigate_path


class PluginsModel:
    """Plugins manaager in the model side"""

    def __init__(self):
        """Initialize plugins manager class"""
        #: str: plugins default path.
        self.plugins_path = os.path.join(
            Path(__file__).resolve().parent.parent, "plugins"
        )

    def load_plugins(self):
        """Load plugins"""
        devices_dict = {}
        plugin_acquisition_modes = {}
        feature_lists_path = os.path.join(get_navigate_path(), "feature_lists")
        if not os.path.exists(feature_lists_path):
            os.makedirs(feature_lists_path)
        feature_list_files = [
            temp
            for temp in os.listdir(feature_lists_path)
            if (temp.endswith(".yml") or temp.endswith(".yaml"))
            and os.path.isfile(os.path.join(feature_lists_path, temp))
        ]
        plugins = os.listdir(self.plugins_path)
        plugins_dict = {}
        # load plugins form plugins folder
        for f in plugins:
            if not os.path.isdir(os.path.join(self.plugins_path, f)):
                continue
            # read "plugin_config.yml"
            plugin_config = load_yaml_file(
                os.path.join(self.plugins_path, f, "plugin_config.yml")
            )
            if plugin_config:
                plugins_dict[plugin_config.get("name", f)] = os.path.join(
                    self.plugins_path, f
                )

        # load plugins from plugins_config
        plugins_config_path = os.path.join(
            get_navigate_path(), "config", "plugins_config.yml"
        )
        if not os.path.exists(plugins_config_path):
            save_yaml_file(
                os.path.join(get_navigate_path(), "config"), {}, "plugins_config.yml"
            )
        else:
            plugins_config = load_yaml_file(plugins_config_path)
            verified_plugins_config = {}
            for plugin_name, plugin_path in plugins_config.items():
                if not plugin_path:
                    print(
                        f"Plugin '{plugin_name}' is not installed correctly. Please reinstall it if necessary."
                    )
                    continue
                if os.path.exists(plugin_path):
                    if plugin_name in plugins_dict:
                        print(
                            f"There are two plugins named '{plugin_name}'. Please rename Plugin '{plugin_name}' in plugins folder!"
                        )
                    plugins_dict[plugin_name] = plugin_path
                    verified_plugins_config[plugin_name] = plugin_path
                else:
                    print(
                        f"Couldn't load plugin '{plugin_name}', please make sure it exits!"
                    )
            save_yaml_file(
                os.path.join(get_navigate_path(), "config"),
                verified_plugins_config,
                "plugins_config.yml",
            )

        for _, plugin_path in plugins_dict.items():
            if not os.path.isdir(plugin_path):
                continue

            # read "plugin_config.yml"
            plugin_config = load_yaml_file(
                os.path.join(plugin_path, "plugin_config.yml")
            )
            if plugin_config is None:
                continue
            plugin_name = plugin_config.get("name", _)

            # feature
            features_dir = os.path.join(plugin_path, "model", "features")
            if os.path.exists(features_dir):
                features = os.listdir(features_dir)
                for feature in features:
                    feature_file = os.path.join(features_dir, feature)
                    if os.path.isfile(feature_file):
                        temp = load_module_from_file(feature, feature_file)
                        for c in dir(temp):
                            if inspect.isclass(getattr(temp, c)):
                                setattr(feature_related_functions, c, getattr(temp, c))
            # feature list
            plugin_feature_list = os.path.join(plugin_path, "feature_list.py")
            if os.path.exists(plugin_feature_list):
                module = load_module_from_file("feature_list_temp", plugin_feature_list)
                features = [
                    f
                    for f in dir(module)
                    if isinstance(getattr(module, f), FeatureList)
                ]
                for feature_name in features:
                    feature = getattr(module, feature_name)
                    feature_list_name = feature.feature_list_name
                    feature_list_file_name = "_".join(feature_list_name.split())
                    feature_list_content = {
                        "module_name": feature_name,
                        "feature_list_name": feature_list_name,
                        "filename": plugin_feature_list,
                    }
                    save_yaml_file(
                        feature_lists_path,
                        feature_list_content,
                        f"{feature_list_file_name}.yml",
                    )

            # acquisition mode
            acquisition_modes = plugin_config.get("acquisition_modes", [])
            for acquisition_mode_config in acquisition_modes:
                acquisition_file = os.path.join(
                    plugin_path, acquisition_mode_config["file_name"]
                )
                if os.path.exists(acquisition_file):
                    module = load_module_from_file(
                        acquisition_mode_config["file_name"][:-3], acquisition_file
                    )
                    acquisition_mode = [
                        m
                        for m in dir(module)
                        if isinstance(getattr(module, m), AcquisitionMode)
                    ]
                    if acquisition_mode:
                        plugin_acquisition_modes[
                            acquisition_mode_config["name"]
                        ] = getattr(module, acquisition_mode[0])(
                            acquisition_mode_config["name"]
                        )

            # device
            device_dir = os.path.join(plugin_path, "model", "devices")
            if os.path.exists(device_dir):
                devices = os.listdir(device_dir)
                for device in devices:
                    device_path = os.path.join(device_dir, device)
                    if not os.path.isdir(device_path):
                        continue
                    try:
                        module = load_module_from_file(
                            "device_module",
                            os.path.join(device_path, "device_startup_functions.py"),
                        )
                    except FileNotFoundError:
                        continue
                    if module:
                        try:
                            device_type_name = module.DEVICE_TYPE_NAME
                        except AttributeError:
                            print(
                                f"Plugin device: {device} is not set correctly!"
                                "Please make sure the DEVICE_TYPE_NAME is given right!"
                            )
                            continue
                        devices_dict[device_type_name] = {}
                        devices_dict[device_type_name][
                            "ref_list"
                        ] = module.DEVICE_REF_LIST
                        devices_dict[device_type_name][
                            "load_device"
                        ] = module.load_device
                        devices_dict[device_type_name][
                            "start_device"
                        ] = module.start_device
        return devices_dict, plugin_acquisition_modes



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.waveforms import camera_exposure

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class DAQBase:
    """DAQBase - Parent class for Data Acquisition (DAQ) classes."""

    def __init__(self, configuration):
        """Initializes the DAQBase class.

        Parameters
        ----------
        configuration : dict
            Dictionary of configuration parameters
        """

        #: dict: Dictionary of configuration parameters
        self.configuration = configuration

        #: dict: Dictionary of waveform constants
        self.waveform_constants = self.configuration["waveform_constants"]

        #: str: Name of the active microscope
        self.microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]

        #: dict: Dictionary of DAQ parameters
        self.daq_parameters = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]

        # Initialize Variables
        #: float: Sample rate of the DAQ
        self.sample_rate = self.daq_parameters["sample_rate"]

        #: dict: Sweep times for different channels
        self.sweep_times = None
        
        #: dict: exposure times for different channels
        self.exposure_times = None

        # Remote Focus Parameters
        #: dict: Dictionary of remote focus ramp falling percentages
        self.remote_focus_ramp_falling = {}
        for m in self.configuration["configuration"]["microscopes"].keys():
            self.remote_focus_ramp_falling[m] = self.configuration["configuration"][
                "microscopes"
            ][m]["remote_focus_device"]["ramp_falling_percent"]

        # Camera Parameters
        #: float: Camera delay percentage
        self.camera_delay_percent = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["camera"]["delay_percent"]

        #: dict: Dictionary of waveforms.
        self.waveform_dict = {}

        #: int: Number of times to repeat the waveform
        self.waveform_repeat_num = 1

        #: int: Number of times to expand the waveform
        self.waveform_expand_num = 1

    def calculate_all_waveforms(self, microscope_name, exposure_times, sweep_times):
        """Pre-calculates all waveforms necessary for the acquisition and organizes in
        a dictionary format.

        Parameters
        ----------
        microscope_name : str
            Name of the active microscope
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel

        Returns
        -------
        self.waveform_dict : dict
            Dictionary of waveforms to pass to galvo and ETL, plus a camera waveform for
            display purposes.
        """
        self.exposure_times = exposure_times
        self.sweep_times = sweep_times
        self.waveform_dict = dict.fromkeys(self.waveform_dict, None)
        self.enable_microscope(microscope_name)

        microscope_state = self.configuration["experiment"]["MicroscopeState"]

        # Iterate through the dictionary.
        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:
                exposure_time = exposure_times[channel_key]
                sweep_time = sweep_times[channel_key]

                self.waveform_dict[channel_key] = camera_exposure(
                    sample_rate=self.sample_rate,
                    sweep_time=sweep_time,
                    exposure=exposure_time,
                    camera_delay=self.camera_delay_percent,
                )

        return self.waveform_dict

    def enable_microscope(self, microscope_name):
        """Enables the microscope.

        Parameters
        ----------
        microscope_name : str
            Name of the active microscope
        """
        if microscope_name != self.microscope_name:
            self.microscope_name = microscope_name

        self.camera_delay_percent = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]["delay_percent"]
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sample_rate"]



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import platform
import tkinter as tk
import logging
import threading

# Third Party Imports
import cv2
from PIL import Image, ImageTk
import matplotlib.pyplot as plt
import numpy as np
import copy

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.model.analysis.camera import compute_signal_to_noise
from navigate.tools.common_functions import VariableWithLock

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class CameraViewController(GUIController):
    """Camera View Controller Class."""

    def __init__(self, view, parent_controller=None):
        """Initialize the Camera View Controller Class.

        Parameters
        ----------
        view : tkinter.Frame
            The tkinter frame that contains the widgets.
        parent_controller : Controller
            The parent controller of the camera view controller.
        """
        super().__init__(view, parent_controller)

        #: SharedNDArray: The shared array that contains the image data.
        self.data_buffer = None

        #: VariableWithLock: The variable that indicates if the image is being
        # displayed.
        self.is_displaying_image = VariableWithLock(bool)

        #: logging.Logger: The logger for the camera view controller.
        self.logger = logging.getLogger(p)

        #: dict: The dictionary of image metrics widgets.
        self.image_metrics = view.image_metrics.get_widgets()

        #: dict: The dictionary of image palette widgets.
        self.image_palette = view.scale_palette.get_widgets()

        #: tkinter.Canvas: The tkinter canvas that displays the image.
        self.canvas = self.view.canvas

        # Bindings for changes to the LUT
        for color in self.image_palette.values():
            color.widget.config(command=self.update_LUT)
        self.update_snr()

        # Binding for adjusting the lookup table min and max counts.
        # keys = ['Autoscale', 'Min','Max']
        self.image_palette["Min"].widget.config(command=self.update_min_max_counts)
        self.image_palette["Max"].widget.config(command=self.update_min_max_counts)
        self.image_palette["Autoscale"].widget.config(
            command=self.toggle_min_max_buttons
        )

        # Transpose and live bindings
        self.image_palette["Flip XY"].widget.config(command=self.transpose_image)
        self.view.live_frame.live.bind(
            "<<ComboboxSelected>>", self.update_display_state
        )

        #: event: The resize event id.
        self.resize_event_id = None
        if platform.system() == "Windows":
            self.resize_event_id = self.view.bind("<Configure>", self.resize)

        #: int: The width of the canvas.
        #: int: The height of the canvas.
        self.width, self.height = 663, 597
        self.canvas_width, self.canvas_height = (
            self.view.canvas_width,
            self.view.canvas_height,
        )
        self.canvas_width, self.canvas_height = 512, 512

        # Right-Click Binding
        #: tkinter.Menu: The tkinter menu that pops up on right click.
        self.menu = tk.Menu(self.canvas, tearoff=0)
        self.menu.add_command(label="Move Here", command=self.move_stage)
        self.menu.add_command(label="Reset Display", command=self.reset_display)
        self.menu.add_command(label="Mark Position", command=self.mark_position)

        #: int: The x position of the mouse.
        self.move_to_x = None

        #: int: The y position of the mouse.
        self.move_to_y = None

        #: numpy.ndarray: The image data.
        self.tk_image = None

        #: numpy.ndarray: The image data.
        self.image = None

        #: numpy.ndarray: The image data.
        self.cross_hair_image = None

        #: numpy.ndarray: The image data.
        self.saturated_pixels = None

        #: numpy.ndarray: The image data.
        self.down_sampled_image = None

        #: numpy.ndarray: The image data.
        self.zoom_image = None

        #: bool: The autoscale flag.
        self.autoscale = True

        #: int: The maximum image counts.
        self.max_counts = None

        #: int: The minimum image counts.
        self.min_counts = None

        #: bool: The crosshair flag.
        self.apply_cross_hair = True

        #: str: The mode of the camera view controller.
        self.mode = "stop"

        #: bool: The transpose flag.
        self.transpose = False

        #: str: The display state.
        self.display_state = "Live"

        # Colormap Information
        #: matplotlib.colors.LinearSegmentedColormap: The colormap.
        self.colormap = plt.get_cmap("gist_gray")

        #: int: The number of images displayed.
        self.image_count = 0

        #: ndarray: A temporary array for image processing.
        self.temp_array = None

        #: int: The number of frames to average.
        self.rolling_frames = 1

        #: list: The list of maximum intensity values.
        self.max_intensity_history = []

        #: int: The bit-depth for PIL presentation.
        self.bit_depth = 8

        #: float: The zoom value for image display.
        self.zoom_value = 1

        #: float: The zoom scale for image display.
        self.zoom_scale = 1

        #: numpy.ndarray: The zoom rectangle.
        self.zoom_rect = np.array([[0, self.canvas_width], [0, self.canvas_height]])

        #: numpy.ndarray: The zoom offset.
        self.zoom_offset = np.array([[0], [0]])

        #: int: The zoom width.
        self.zoom_width = self.canvas_width

        #: int: The zoom height.
        self.zoom_height = self.canvas_height

        #: int: The canvas width scaling factor.
        self.canvas_width_scale = 4

        #: int: The canvas height scaling factor.
        self.canvas_height_scale = 4

        #: int: The original image height.
        self.original_image_height = 2014

        #: int: The original image width.
        self.original_image_width = 2014

        #: int: The number of slices.
        self.number_of_slices = 0

        #: numpy.ndarray: The image volume.
        self.image_volume = None

        #: int: The total number of images per volume.
        self.total_images_per_volume = 0

        #: int: The number of channels.
        self.number_of_channels = 0

        #: int: The image counter.
        self.image_counter = 0

        #: int: The slice index.
        self.slice_index = 0

        #: int: The channel index.
        self.channel_index = 0

        #: int: The crosshair x position.
        self.crosshair_x = None

        #: int: The crosshair y position.
        self.crosshair_y = None

        #: bool: The display mask flag.
        self.mask_color_table = None

        #: bool: Whether or not to flip the image.
        self.flip_flags = None

        #: bool: Image catche flag
        self.image_catche_flag = True

        # ilastik mask
        #: bool: The display mask flag for ilastik.
        self.display_mask_flag = False

        #: threading.Lock: The lock for the ilastik mask.
        self.ilastik_mask_ready_lock = threading.Lock()

        #: numpy.ndarray: The ilastik mask.
        self.ilastik_seg_mask = None

    def update_snr(self):
        """Updates the signal to noise ratio."""
        #: bool: The signal to noise ratio flag.
        self._snr_selected = False
        #: numpy.ndarray: The offset of the image.
        #: numpy.ndarray: The variance of the image.
        self._offset, self._variance = None, None
        off, var = self.parent_controller.model.get_offset_variance_maps()
        if off is None:
            self.image_palette["SNR"].grid_remove()
        else:
            self._offset, self._variance = copy.deepcopy(off), copy.deepcopy(var)
            self.image_palette["SNR"].grid(row=3, column=0, sticky=tk.NSEW, pady=3)

    def slider_update(self, event):
        """Updates the image when the slider is moved.

        Parameters
        ----------
        event : tkinter event
            The tkinter event that triggered the function.
        """

        slider_index = self.view.slider.get()
        channel_display_index = 0
        self.retrieve_image_slice_from_volume(
            slider_index=slider_index, channel_display_index=channel_display_index
        )
        self.reset_display()

    def update_display_state(self, event):
        """Image Display Combobox Called.

        Sets self.display_state to desired display format.
        Toggles state of slider widget.
        Sets number of positions.

        Parameters
        ----------
        event : tkinter event
            The tkinter event that triggered the function.
        """
        self.display_state = self.view.live_frame.live.get()
        # Slice in the XY Dimension.
        if self.display_state == "XY Slice":
            print("XY Slice")
            try:
                slider_length = np.shape(self.image_volume)[2] - 1
            except IndexError:
                slider_length = (
                    self.parent_controller.configuration["experiment"][
                        "MicroscopeState"
                    ]["number_z_steps"]
                    - 1
                )
        if self.display_state == "YZ Slice":
            try:
                slider_length = np.shape(self.image_volume)[0] - 1
            except IndexError:
                slider_length = (
                    self.parent_controller.configuration["experiment"][
                        "CameraParameters"
                    ]["y_pixels"]
                    - 1
                )
        if self.display_state == "YZ Slice":
            try:
                slider_length = np.shape(self.image_volume)[1] - 1
            except IndexError:
                slider_length = (
                    self.parent_controller.configuration["experiment"][
                        "CameraParameters"
                    ]["x_pixels"]
                    - 1
                )

        if self.display_state.find("Slice") != -1:
            self.view.slider.slider_widget.configure(
                to=slider_length, tickinterval=(slider_length / 5), state="normal"
            )
        else:
            self.view.slider.slider_widget.configure(state="disabled")

    def get_absolute_position(self):
        """Gets the absolute position of the computer mouse.

        Returns
        -------
        x : int
            The x position of the mouse.
        y : int
            The y position of the mouse.

        Examples
        --------
        >>> x, y = self.get_absolute_position()
        """
        x = self.parent_controller.view.winfo_pointerx()
        y = self.parent_controller.view.winfo_pointery()
        return x, y

    def popup_menu(self, event):
        """Right-Click Popup Menu

        Parameters
        ----------
        event : tkinter.Event
            x, y location.  0,0 is top left corner.

        """
        try:
            # only popup the menu when click on image
            if event.x >= self.canvas_width or event.y >= self.canvas_height:
                return
            self.move_to_x = event.x
            self.move_to_y = event.y
            x, y = self.get_absolute_position()
            self.menu.tk_popup(x, y)
        finally:
            self.menu.grab_release()

    def initialize(self, name, data):
        """Sets widgets based on data given from main controller/config.

        Parameters
        ----------
        name : str
            'minmax', 'image'.
        data : list
            Min and max intensity values.

        Examples
        --------
        >>> self.initialize('minmax', [0, 255])
        """
        # Pallete section (colors, autoscale, min/max counts)
        # keys = ['Frames to Avg', 'Image Max Counts', 'Channel']
        if name == "minmax":
            min = data[0]
            max = data[1]

            # Invoking defaults
            self.image_palette["Gray"].widget.invoke()
            self.image_palette["Autoscale"].widget.invoke()

            # Populating defaults
            self.image_palette["Min"].set(min)
            self.image_palette["Max"].set(max)
            self.image_palette["Min"].widget["state"] = "disabled"
            self.image_palette["Max"].widget["state"] = "disabled"

        self.image_palette["Flip XY"].widget.invoke()

        # Image Metrics section
        if name == "image":
            frames = data[0]
            # Populating defaults
            self.image_metrics["Frames"].set(frames)

    #  Set mode for the execute statement in main controller

    def set_mode(self, mode=""):
        """Sets mode of camera_view_controller.

        Parameters
        ----------
        mode : str
            camera_view_controller mode.

        Examples
        --------
        >>> self.set_mode('live')
        """
        self.mode = mode
        if mode == "live" or mode == "stop":
            self.menu.entryconfig("Move Here", state="normal")
        else:
            self.menu.entryconfig("Move Here", state="disabled")

    def mark_position(self):
        """Marks the current position of the microscope in
        the multi-position acquisition table."""
        offset_x, offset_y = self.calculate_offset()
        stage_position = self.parent_controller.execute("get_stage_position")
        if stage_position is not None:
            stage_flip_flags = (
                self.parent_controller.configuration_controller.stage_flip_flags
            )
            stage_position["x"] += offset_x * (-1 if stage_flip_flags["x"] else 1)
            stage_position["y"] -= offset_y * (-1 if stage_flip_flags["y"] else 1)

            # Place the stage position in the multi-position table.
            self.parent_controller.execute("mark_position", stage_position)

    def calculate_offset(self):
        """Calculates the offset of the image.

        Returns
        -------
        offset_x : int
            The offset of the image in x.
        offset_y : int
            The offset of the image in y.
        """
        current_center_x = (self.zoom_rect[0][0] + self.zoom_rect[0][1]) / 2
        current_center_y = (self.zoom_rect[1][0] + self.zoom_rect[1][1]) / 2

        microscope_name = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]["microscope_name"]
        zoom_value = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]["zoom"]
        pixel_size = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][microscope_name]["zoom"]["pixel_size"][zoom_value]

        offset_x = int(
            (self.move_to_x - current_center_x)
            / self.zoom_scale
            * self.canvas_width_scale
            * pixel_size
        )
        offset_y = int(
            (self.move_to_y - current_center_y)
            / self.zoom_scale
            * self.canvas_height_scale
            * pixel_size
        )

        return offset_x, offset_y

    def move_stage(self):
        """Move the stage according to the position the user clicked."""
        offset_x, offset_y = self.calculate_offset()

        self.show_verbose_info(
            f"Try moving stage by {offset_x} in x and {offset_y} in y"
        )

        stage_position = self.parent_controller.execute("get_stage_position")

        if stage_position is not None:
            # TODO: if show image as what the camera gets(flipped one), the stage
            # moving direction should be decided by stage_flip_flags
            # and camera_flip_flags
            stage_flip_flags = (
                self.parent_controller.configuration_controller.stage_flip_flags
            )
            stage_position["x"] += offset_x * (-1 if stage_flip_flags["x"] else 1)
            stage_position["y"] -= offset_y * (-1 if stage_flip_flags["y"] else 1)
            if self.mode == "stop":
                command = "move_stage_and_acquire_image"
            else:
                command = "move_stage_and_update_info"
            self.parent_controller.execute(command, stage_position)
        else:
            tk.messagebox.showerror(
                title="Warning", message="Can't move to there! Invalid stage position!"
            )

    def reset_display(self, display_flag=True):
        """Set the display back to the original digital zoom.

        Parameters
        ----------
        display_flag : bool
            True to display the image, False to not display the image.

        Examples
        --------
        >>> self.reset_display()
        """
        self.zoom_width = self.canvas_width
        self.zoom_height = self.canvas_height
        self.zoom_rect = np.array([[0, self.zoom_width], [0, self.zoom_height]])
        self.zoom_offset = np.array([[0], [0]])
        self.zoom_value = 1
        self.zoom_scale = 1
        if display_flag:
            self.process_image()

    def process_image(self):
        """Process the image to be displayed.

        Applies digital zoom, detects saturation, down-samples the image, scales the
        image intensity, adds a crosshair, applies the lookup table, and populates the
        image.

        Examples
        --------
        >>> self.process_image()
        """
        # self.image -> self.zoom_image.
        self.digital_zoom()

        # self.zoom_image -> self.zoom_image
        self.detect_saturation()

        # self.zoom_image -> self.down_sampled_image
        self.down_sample_image()

        # self.down_sampled_image  -> self.down_sampled_image
        self.scale_image_intensity()

        # self_down_sampled_image -> self.cross_hair_image
        self.add_crosshair()

        # self_cross_hair_image -> self.cross_hair_image)
        self.apply_LUT()

        # self.cross_hair_image -> display...
        self.populate_image()

    def mouse_wheel(self, event):
        """Digitally zooms in or out on the image upon scroll wheel event.

        Sets the self.zoom_value between 0.05 and 1 in .05 unit steps.

        Parameters
        ----------
        event : tkinter.Event
            num = 4 is zoom out.
            num = 5 is zoom in.
            x, y location.  0,0 is top left corner.

        Examples
        --------
        >>> self.mouse_wheel(event)
        """
        if event.x >= self.canvas_width or event.y >= self.canvas_height:
            return
        self.zoom_offset = np.array([[int(event.x)], [int(event.y)]])
        delta = 120 if platform.system() != "Darwin" else 1
        threshold = event.delta / delta
        if (event.num == 4) or (threshold > 0):
            # Zoom out event.
            self.zoom_value = 0.95
        if (event.num == 5) or (threshold < 0):
            # Zoom in event.
            self.zoom_value = 1.05

        self.zoom_scale *= self.zoom_value
        self.zoom_width /= self.zoom_value
        self.zoom_height /= self.zoom_value

        if self.zoom_width > self.canvas_width or self.zoom_height > self.canvas_height:
            self.reset_display(False)
        elif self.zoom_width < 5 or self.zoom_height < 5:
            return

        self.process_image()

    def digital_zoom(self):
        """Apply digital zoom.

        The x and y positions are between 0
        and the canvas width and height respectively.

        """
        self.zoom_rect = self.zoom_rect - self.zoom_offset
        self.zoom_rect = self.zoom_rect * self.zoom_value
        self.zoom_rect = self.zoom_rect + self.zoom_offset
        self.zoom_offset.fill(0)
        self.zoom_value = 1

        if self.zoom_rect[0][0] > 0 or self.zoom_rect[1][0] > 0:
            self.reset_display(False)

        x_start_index = int(-self.zoom_rect[0][0] / self.zoom_scale)
        x_end_index = int(x_start_index + self.zoom_width)
        y_start_index = int(-self.zoom_rect[1][0] / self.zoom_scale)
        y_end_index = int(y_start_index + self.zoom_height)

        # crosshair
        crosshair_x = (self.zoom_rect[0][0] + self.zoom_rect[0][1]) / 2
        crosshair_y = (self.zoom_rect[1][0] + self.zoom_rect[1][1]) / 2
        if crosshair_x < 0 or crosshair_x >= self.canvas_width:
            crosshair_x = -1
        if crosshair_y < 0 or crosshair_y >= self.canvas_height:
            crosshair_y = -1
        self.crosshair_x = int(crosshair_x)
        self.crosshair_y = int(crosshair_y)

        self.zoom_image = self.image[
            int(y_start_index * self.canvas_height_scale) : int(
                y_end_index * self.canvas_height_scale
            ),
            int(x_start_index * self.canvas_width_scale) : int(
                x_end_index * self.canvas_width_scale
            ),
        ]

    def left_click(self, event):
        """Toggles cross-hair on image upon left click event.

        Parameters
        ----------
        event : tkinter.Event
            Tkinter event.
        """
        if self.image is not None:
            # If True, make False. If False, make True.
            self.apply_cross_hair = not self.apply_cross_hair
            self.add_crosshair()
            self.apply_LUT()
            self.populate_image()

    def update_max_counts(self):
        """Update the max counts in the camera view.

        Function gets the number of frames to average from the VIEW.

        If frames to average == 0 or 1, provides the maximum value from the last
        acquired data.
        """

        # If the array is larger than 32 entries, remove the 0th entry.
        if len(self.max_intensity_history) > (2**5):
            self.max_intensity_history = self.max_intensity_history[1:]

        # Get the number of frames to average from the VIEW
        self.rolling_frames = int(self.image_metrics["Frames"].get())

        # Make sure the array is longer than the number of frames to average.
        if self.rolling_frames > len(self.max_intensity_history):
            self.rolling_frames = len(self.max_intensity_history)
            self.image_metrics["Frames"].set(self.rolling_frames)

        if self.rolling_frames == 0:
            # Cannot average 0 frames. Set to 1, and report max intensity
            self.image_metrics["Frames"].set(1)
            self.image_metrics["Image"].set(f"{self.max_intensity_history[-1]:.0f}")
        elif self.rolling_frames == 1:
            self.image_metrics["Image"].set(f"{self.max_intensity_history[-1]:.0f}")
        elif self.rolling_frames > 1:
            rolling_average = (
                sum(self.max_intensity_history[-self.rolling_frames :])
                / self.rolling_frames
            )
            self.image_metrics["Image"].set(f"{rolling_average:.0f}")

    def down_sample_image(self):
        """Down-sample the data for image display according to widget size."""
        sx, sy = self.canvas_width, self.canvas_height
        self.down_sampled_image = cv2.resize(self.zoom_image, (sx, sy))

    def scale_image_intensity(self):
        """Scale the data to the min/max counts, and adjust bit-depth."""
        if self.autoscale is True:
            self.max_counts = np.max(self.down_sampled_image)
            self.min_counts = np.min(self.down_sampled_image)
        else:
            self.update_min_max_counts()

        scaling_factor = 1
        self.down_sampled_image = scaling_factor * (
            (self.down_sampled_image - self.min_counts)
            / (self.max_counts - self.min_counts)
        )
        self.down_sampled_image[self.down_sampled_image < 0] = 0
        self.down_sampled_image[
            self.down_sampled_image > scaling_factor
        ] = scaling_factor

    def populate_image(self):
        """Converts image to an ImageTk.PhotoImage and populates the Tk Canvas"""
        if self.display_mask_flag:
            self.ilastik_mask_ready_lock.acquire()
            temp_img1 = self.cross_hair_image.astype(np.uint8)
            img1 = Image.fromarray(temp_img1)
            temp_img2 = cv2.resize(self.ilastik_seg_mask, temp_img1.shape[:2])
            img2 = Image.fromarray(temp_img2)
            temp_img = Image.blend(img1, img2, 0.2)
        else:
            temp_img = Image.fromarray(self.cross_hair_image.astype(np.uint8))
        
        # when calling ImageTk.PhotoImage() to generate a new image, it will destroy
        # what the canvas is showing and cause a blink.
        if self.image_catche_flag:
            self.tk_image = ImageTk.PhotoImage(temp_img)
            self.canvas.create_image(0, 0, image=self.tk_image, anchor="nw")
        else:
            self.tk_image2 = ImageTk.PhotoImage(temp_img)
            self.canvas.create_image(0, 0, image=self.tk_image2, anchor="nw")

        self.image_catche_flag = not self.image_catche_flag

    def initialize_non_live_display(self, buffer, microscope_state, camera_parameters):
        """Initialize the non-live display.

        Starts image and slice counter,
        number of channels,
        number of slices,
        images per volume,
        and image volume.

        Parameters
        ----------
        buffer : numpy.ndarray
            Image data.
        microscope_state : dict
            Microscope state.
        camera_parameters : dict
            Camera parameters.

        Example
        -------
        >>> self.initialize_non_live_display(buffer,
        >>> microscope_state, camera_parameters)
        """
        self.data_buffer = buffer
        self.is_displaying_image.value = False
        self.image_counter = 0
        self.slice_index = 0
        self.number_of_channels = len(microscope_state["channels"])
        self.number_of_slices = int(microscope_state["number_z_steps"])
        self.total_images_per_volume = self.number_of_channels * self.number_of_slices
        self.original_image_width = int(camera_parameters["x_pixels"])
        self.original_image_height = int(camera_parameters["y_pixels"])
        self.flip_flags = (
            self.parent_controller.configuration_controller.camera_flip_flags
        )

        self.update_canvas_size()

        self.reset_display(False)

    def identify_channel_index_and_slice(self, microscope_state, images_received):
        """As images arrive, identify channel index and slice.

        Parameters
        ----------
        microscope_state : dict
            State of the microscope
        images_received : int
            Number of images received.

        Example
        -------
        >>> self.identify_channel_index_and_slice(microscope_state, images_received)
        """
        # Reset the image counter after the full acquisition of an image volume.
        if self.image_counter == self.total_images_per_volume:
            self.image_counter = 0

        # Store each image to the pre-allocated memory.
        if microscope_state["stack_cycling_mode"] == "per_stack":
            if (
                0 * self.number_of_slices
                <= self.image_counter
                < 1 * self.number_of_slices
            ):
                self.channel_index = 0
            elif (
                1 * self.number_of_slices
                <= self.image_counter
                < 2 * self.number_of_slices
            ):
                self.channel_index = 1
            elif (
                2 * self.number_of_slices
                <= self.image_counter
                < 3 * self.number_of_slices
            ):
                self.channel_index = 2
            elif (
                3 * self.number_of_slices
                <= self.image_counter
                < 4 * self.number_of_slices
            ):
                self.channel_index = 3
            elif (
                4 * self.number_of_slices
                <= self.image_counter
                < 5 * self.number_of_slices
            ):
                self.channel_index = 4
            else:
                self.channel_index = 0
                print(
                    "Camera View Controller - "
                    "Cannot identify proper channel for per_stack imaging mode."
                )

            self.slice_index = self.image_counter - (
                self.channel_index * self.number_of_slices
            )
            self.image_counter += 1

        elif microscope_state["stack_cycling_mode"] == "per_z":
            # Every image that comes in will be the next channel.
            self.channel_index = images_received % self.number_of_channels
            self.image_volume[:, :, self.slice_index, self.channel_index] = self.image
            if self.channel_index == (self.number_of_channels - 1):
                self.slice_index += 1
            if self.slice_index == self.total_images_per_volume:
                self.slice_index = 0

        # print(self.channel_index, self.slice_index)

    def retrieve_image_slice_from_volume(self, slider_index, channel_display_index):
        """Retrieve image slice from volume.

        Parameters
        ----------
        slider_index : int
            Index of the slider.
        channel_display_index : int
            Index of the channel to display.

        Example
        -------
        >>> self.retrieve_image_slice_from_volume(slider_index, channel_display_index)
        """
        if self.display_state == "XY MIP":
            self.image = np.max(
                self.image_volume[:, :, :, channel_display_index], axis=2
            )
        if self.display_state == "YZ MIP":
            self.image = np.max(
                self.image_volume[:, :, :, channel_display_index], axis=0
            )
        if self.display_state == "ZY MIP":
            self.image = np.max(
                self.image_volume[:, :, :, channel_display_index], axis=1
            )
        if self.display_state == "XY Slice":
            self.image = self.image_volume[:, :, slider_index, channel_display_index]
        if self.display_state == "YZ Slice":
            self.image = self.image_volume[slider_index, :, :, channel_display_index]
        if self.display_state == "ZY Slice":
            self.image = self.image_volume[:, slider_index, :, channel_display_index]

    def display_image(self, image_id):
        """Display an image using the LUT specified in the View.

        If Autoscale is selected, automatically calculates
        the min and max values for the data.

        If Autoscale is not selected, takes the user values
        as specified in the min and max counts.

        Parameters
        ----------
        image_id: int
            frame index in the data_buffer.

        Example
        -------
        >>> self.display_image(image_id)
        """

        # Identify image identity (e.g., slice #, channel #).
        # self.identify_channel_index_and_slice(microscope_state=microscope_state,
        #                                       images_received=images_received)

        # Place image in memory
        # TODO: This is the slow part
        # self.image_volume[:, :, self.slice_index,
        # self.channel_index] = image[:, ] # copy

        # Store the maximum intensity value for the image.
        image = self.data_buffer[image_id]
        # flip back image
        if self.flip_flags["x"] and self.flip_flags["y"]:
            image = image[::-1, ::-1]
        elif self.flip_flags["x"]:
            image = image[:, ::-1]
        elif self.flip_flags["y"]:
            image = image[::-1, :]

        self.max_intensity_history.append(np.max(image))

        # If the user has toggled the transpose button, transpose the image.
        if self.transpose:
            self.image = image.T
        else:
            self.image = image

        if self._snr_selected:
            self.image = compute_signal_to_noise(
                self.image, self._offset, self._variance
            )

        # MIP and Slice Mode TODO: Consider channels
        # if self.display_state != 'Live':
        #     slider_index = self.view.slider.slider_widget.get()
        #     channel_display_index = 0
        #     self.retrieve_image_slice_from_volume(slider_index=slider_index,
        #                                           channel_display_index=channel_display_index)
        #
        # else:
        self.process_image()
        self.update_max_counts()
        self.image_metrics["Channel"].set(self.channel_index)
        self.image_count = self.image_count + 1
        with self.is_displaying_image as is_displaying_image:
            is_displaying_image.value = False

    def add_crosshair(self):
        """Adds a cross-hair to the image."""
        self.cross_hair_image = np.copy(self.down_sampled_image)
        if self.apply_cross_hair:
            self.cross_hair_image[:, self.crosshair_x] = 1
            self.cross_hair_image[self.crosshair_y, :] = 1

    def apply_LUT(self):
        """Applies a LUT to an image.

        Red is reserved for saturated pixels.
        self.color_values = ['gray', 'gradient', 'rainbow']
        """
        # if self.colormap == 'gradient':
        #     self.cross_hair_image = self.rainbow_lut(self.cross_hair_image)
        # elif self.colormap == 'rainbow':
        #     self.cross_hair_image = self.gradient_lut(self.cross_hair_image)
        # elif self.colormap == 'RdBu_r':
        #     self.cross_hair_image = self.rdbu_r_lut(self.cross_hair_image)
        # else:
        #     self.cross_hair_image = self.gray_lut(self.cross_hair_image)
        self.cross_hair_image = self.colormap(self.cross_hair_image)

        # Convert RGBA to RGB Image.
        self.cross_hair_image = self.cross_hair_image[:, :, :3]

        # Specify the saturated values in the red channel
        if np.any(self.saturated_pixels):
            # Saturated pixels is an array of True or
            # False statements same size as the image.

            # Pull out the red image from the RGBA
            # Set saturated pixels to 1, put back into array.
            red_image = self.cross_hair_image[:, :, 2]
            red_image[self.saturated_pixels] = 1
            self.cross_hair_image[:, :, 2] = red_image

        # Scale back to an 8-bit image.
        self.cross_hair_image = self.cross_hair_image * (2**self.bit_depth - 1)

    def update_LUT(self):
        """Update the LUT in the Camera View.

        When the LUT is changed in the GUI, this function is called.
        Updates the LUT.

        Parameters
        ----------
        self.image : np.array
            Must be a 2D image.

        Returns
        -------
        self.apply_LUT_image : np.arrays

        Example
        -------
        >>> self.update_LUT()
        """
        if self.image is None:
            pass
        else:
            cmap_name = self.view.scale_palette.color.get()
            self._snr_selected = (
                True if cmap_name == "RdBu_r" else False
            )  # TODO: Don't use a proxy for SNR
            self.colormap = plt.get_cmap(cmap_name)
            self.add_crosshair()
            self.apply_LUT()
            self.populate_image()
            logger.debug(f"Updating the LUT, {cmap_name}")

    def detect_saturation(self):
        """Look for any pixels at the maximum intensity allowable for the camera.

        Parameters
        ----------
        self.image : np.array
            Must be a 2D image.

        Returns
        -------
        self.saturated_pixels : np.array
            Boolean array of the same size as the image.
        """
        saturation_value = 2**16 - 1
        self.saturated_pixels = self.zoom_image[self.zoom_image > saturation_value]

    def toggle_min_max_buttons(self):
        """Checks the value of the autoscale widget.

        If enabled, the min and max widgets are disabled and the image intensity is
        autoscaled. If disabled, miu and max widgets are enabled, and image intensity
        scaled.
        """
        self.autoscale = self.image_palette["Autoscale"].get()

        if self.autoscale is True:  # Autoscale Enabled
            self.image_palette["Min"].widget["state"] = "disabled"
            self.image_palette["Max"].widget["state"] = "disabled"
            logger.info("Autoscale Enabled")

        elif self.autoscale is False:  # Autoscale Disabled
            self.image_palette["Min"].widget["state"] = "normal"
            self.image_palette["Max"].widget["state"] = "normal"
            logger.info("Autoscale Disabled")
            self.update_min_max_counts()

    def transpose_image(self):
        """Get Flip XY widget value from the View.

        If True, transpose the image.

        Returns
        -------
        self.image : np.array
            Transposed image.
        """
        self.transpose = self.image_palette["Flip XY"].get()

    def update_min_max_counts(self):
        """Get min and max count values from the View.

        When the min and max counts are toggled in the GUI, this function is called.
        Updates the min and max values.

        Returns
        -------
        self.min_counts : int
            Minimum counts for the image.
        self.max_counts : int
            Maximum counts for the image.

        Example
        -------
        >>> self.update_min_max_counts()
        """
        if self.image_palette["Min"].get() != "":
            self.min_counts = float(self.image_palette["Min"].get())
        if self.image_palette["Max"].get() != "":
            self.max_counts = float(self.image_palette["Max"].get())
        logger.debug(
            f"Min and Max counts scaled to, {self.min_counts}, {self.max_counts}"
        )

    def set_mask_color_table(self, colors):
        """Set up segmentation mask color table

        Parameters
        ----------
        colors : list
            List of colors to use for the segmentation mask

        Returns
        -------
        self.mask_color_table : np.array
            Array of colors to use for the segmentation mask

        Example
        -------
        >>> self.set_mask_color_table()
        """
        self.mask_color_table = np.zeros((256, 1, 3), dtype=np.uint8)
        self.mask_color_table[0] = [0, 0, 0]
        for i in range(len(colors)):
            color_hex = colors[i]
            self.mask_color_table[i + 1] = [
                int(color_hex[1:3], 16),
                int(color_hex[3:5], 16),
                int(color_hex[5:], 16),
            ]
        if not self.ilastik_mask_ready_lock.locked():
            self.ilastik_mask_ready_lock.acquire()

    def display_mask(self, mask):
        """Display segmentation mask

        Parameters
        ----------
        mask : np.array
            Segmentation mask to display

        Example
        -------
        >>> self.display_mask()
        """
        self.ilastik_seg_mask = cv2.applyColorMap(mask, self.mask_color_table)
        self.ilastik_mask_ready_lock.release()

    def resize(self, event):
        """Resize the window.

        Parameters
        ----------
        event : tkinter.Event
            Tkinter event.
        """
        if self.view.is_popup is False and event.widget != self.view:
            return
        if self.view.is_popup is True and event.widget.widgetName != "toplevel":
            return
        if self.resize_event_id:
            self.view.after_cancel(self.resize_event_id)
        self.resize_event_id = self.view.after(
            1000, lambda: self.refresh(event.width, event.height)
        )

    def refresh(self, width, height):
        """Refresh the window.

        Parameters
        ----------
        width : int
            Width of the window.
        height : int
            Height of the window.
        """
        if width == self.width and height == self.height:
            return
        self.canvas_width = width - self.view.image_metrics.winfo_width() - 24
        self.canvas_height = height - 85
        self.view.canvas.config(width=self.canvas_width, height=self.canvas_height)
        self.view.update_idletasks()

        if self.view.is_popup:
            self.width, self.height = self.view.winfo_width(), self.view.winfo_height()
        else:
            self.width, self.height = width, height

        # if resize the window during acquisition, the image showing should be updated
        self.update_canvas_size()
        self.reset_display(False)

    def update_canvas_size(self):
        """Update the canvas size."""
        r_canvas_width = int(self.view.canvas["width"])
        r_canvas_height = int(self.view.canvas["height"])
        img_ratio = self.original_image_width / self.original_image_height
        canvas_ratio = r_canvas_width / r_canvas_height

        if canvas_ratio > img_ratio:
            self.canvas_height = r_canvas_height
            self.canvas_width = int(r_canvas_height * img_ratio)
        else:
            self.canvas_width = r_canvas_width
            self.canvas_height = int(r_canvas_width / img_ratio)

        self.canvas_width_scale = float(self.original_image_width / self.canvas_width)
        self.canvas_height_scale = float(
            self.original_image_height / self.canvas_height
        )

    def try_to_display_image(self, image_id):
        """Try to display an image.

        Parameters
        ----------
        image_id : int
            Frame index in the data_buffer.
        """
        with self.is_displaying_image as is_displaying_image:
            if is_displaying_image.value:
                return
            is_displaying_image.value = True

        display_thread = threading.Thread(target=self.display_image, args=(image_id,))
        display_thread.start()
